from __future__ import division, absolute_import, print_function

import inspect
from abc import ABCMeta, abstractmethod, abstractproperty
from six import add_metaclass
from itertools import chain

import numpy as np
import tensorflow as tf
from tensorflow.python.ops import rnn
from tensorflow.python.ops import init_ops

from odin import backend as K
from odin.autoconfig import randint
from odin.nnet.base import NNOp
from odin.nnet.normalization import BatchNorm
from odin.backend.role import (InitialState, Weight, Bias, Parameter,
                               has_roles, BatchNormShiftParameter,
                               BatchNormScaleParameter,
                               BatchNormPopulationMean,
                               BatchNormPopulationInvStd)
from odin.utils import (as_tuple, is_string, is_number, wprint)

# ===========================================================================
# Helper
# ===========================================================================
def _init_input2hidden(ops, rnn_mode, input_mode, W_init, input_dims, hidden_dims):
  # N represent the number of gates
  if 'rnn' in rnn_mode:
    N = 1
    msg = '(W_hid)'
  elif rnn_mode == 'gru':
    N = 3
    msg = '(W_input_to_updategate, W_input_to_resetgate, W_input_to_hiddenupdate)'
  elif rnn_mode == 'lstm':
    N = 4
    msg = '(W_input_to_inputgate, W_input_to_forgetgate, W_input_to_hidden, W_input_to_outputgate)'
  # ====== check input ====== #
  if input_mode != 'skip':
    ops.get_variable_nnop(initializer=W_init, shape=(input_dims, hidden_dims * N),
                     name='W_in', roles=Weight)
    if input_mode == 'norm':
      ops.get_variable_nnop(initializer=init_ops.constant_initializer(0.), shape=(hidden_dims * N,),
                            name='beta', roles=BatchNormShiftParameter)
      ops.get_variable_nnop(initializer=init_ops.constant_initializer(1.), shape=(hidden_dims * N,),
                            name='gamma', roles=BatchNormScaleParameter)
      ops.get_variable_nnop(initializer=init_ops.constant_initializer(0.), shape=(hidden_dims * N,),
                            name='mean', roles=BatchNormPopulationMean)
      ops.get_variable_nnop(initializer=init_ops.constant_initializer(1.), shape=(hidden_dims * N,),
                            name='inv_std', roles=BatchNormPopulationInvStd)
  # skip input mode
  elif input_dims != hidden_dims and \
  input_dims != hidden_dims * N: # 3 gates + 1 hid_update
    raise Exception('Skip input mode, input trailing_dimension=%d '
                    '(the final dim) must equal to the number of hidden '
                    'units (tied input connection), or %d-th the number '
                    'of hidden units = %d, which include: ' + msg %
                    (input_dims, N, hidden_dims * N))


def _check_cudnn_hidden_init(s0, shape, nnops, name):
  nb_layers, batch_size, hidden_size = shape
  # ====== init s0 ====== #
  if s0 is None and hasattr(nnops, name):
    s0 = getattr(nnops, name)
  elif s0 is not None:
    if hasattr(s0, '__call__') or K.is_variable(s0) or \
    isinstance(s0, np.ndarray):
      _ = (nb_layers, 1, hidden_size) \
          if hasattr(s0, '__call__') or isinstance(s0, np.ndarray) \
          else s0.shape
      s0 = nnops.config.create_params(
          s0, shape=_, name=name, roles=InitialState)
    # ====== check s0 shape ====== #
    init_shape = s0.shape
    if s0.shape.ndims == 2:
      if s0.shape[-1].value != hidden_size:
        raise ValueError('init state has %d dimension, but the hidden_size=%d' %
                        (init_shape[-1], hidden_size))
    elif init_shape[::2] != (nb_layers, hidden_size):
      raise ValueError('Require init states of size: %s, but '
                       'given state of size: %s' % (shape, init_shape))
    # ====== return the right shape ====== #
    setattr(nnops, name, s0)
  return s0


# ===========================================================================
# Dynamic RNN
# ===========================================================================
def _infer_variable_role(variables):
  for v in variables:
    name = v.name.split(':')[0].split("/")[-1]
    if 'kernel' in name:
      K.role.add_roles(v, K.role.Weight)
    elif 'bias' in name:
      K.role.add_roles(v, K.role.Bias)
    elif '_w' in name:
      K.role.add_roles(v, K.role.Weight)
    elif '_v' in name:
      K.role.add_roles(v, K.role.Weight)
    else:
      raise ValueError("Unknown role for variable with name: " + name)

def get_cell_info(cell):
  from tensorflow.contrib.rnn.python.ops import rnn_cell
  found_cell = None
  for key, val in inspect.getmembers(K.rnn_cell) + inspect.getmembers(rnn_cell):
    if inspect.isclass(val) and issubclass(val, tf.contrib.rnn.RNNCell):
      if (isinstance(cell, str) and (key == cell or str(val) == cell)) \
      or cell == val:
        found_cell = val; break
  # ====== get cell init info ====== #
  sign = inspect.signature(found_cell.__init__)
  args = []
  kwargs = {}
  for n, p in sign.parameters.items():
    if p.kind in (inspect.Parameter.VAR_POSITIONAL,
                  inspect.Parameter.VAR_KEYWORD) and\
       n != 'self':
      continue
    args.append(n)
    if p.default != inspect.Parameter.empty:
      kwargs[n] = p.default
  return found_cell, args, kwargs

class RNN(NNOp):
  """
  Parameter
  ---------
  attention: None or `tf.contrib.seq2seq.AttentionMechanism`
      two basic attentions: LuongAttention, BahdanauAttention
      or `tf.contrib.rnn.AttentionCellWrapper`
  """

  def __init__(self, cell_type, kwargs={},
               attention=None, attention_kwargs={},
               num_layers=1, bidirectional=False, dynamic=True,
               return_states=False, name=None):
    super(RNN, self).__init__(name=name)
    self._is_initialized_variables = False
    self.cell_type, _args, _kwargs = get_cell_info(cell_type)
    if self.cell_type is None:
      raise RuntimeError("Cannot find any RNNCell with given description: %s"
          % str(cell_type))
    _kwargs.update(kwargs)
    if len(_kwargs) != len(_args):
      raise RuntimeError("Missing following arguments: %s for the RNNCell "
          "of type: %s" % (list(set(_args) - set(_kwargs.keys())),
                           self.cell_type.__name__))
    self.cell_kwargs = _kwargs
    self.num_layers = int(num_layers)
    self.bidirectional = bidirectional
    self.dynamic = dynamic
    self.return_states = return_states
    # ====== attention ====== #
    if attention is True:
      attention = tf.contrib.rnn.AttentionCellWrapper
    if attention is not None and attention is not False and \
    (not issubclass(attention, tf.contrib.seq2seq.AttentionMechanism) and
     attention is not tf.contrib.rnn.AttentionCellWrapper and
     attention is not K.rnn_cell.AttentionCell):
      raise ValueError("`attention` argument must be `None` or instance "
          "of `tensorflow.contrib.seq2seq.AttentionMechanism`.")
    self.attention = attention
    self.attention_kwargs = dict(attention_kwargs)
    # tf.contrib.seq2seq.embedding_attention_seq2seq

  # ==================== Helper ==================== #
  def __cell_creator(self):
    if self.num_layers > 1:
      c = tf.contrib.rnn.MultiRNNCell([self.cell_type(**self.cell_kwargs)
                                       for _ in range(self.num_layers)])
    else:
      c = self.cell_type(**self.cell_kwargs)
    return c

  def __attention_creator(self, cell, X, memory):
    kwargs = self.attention_kwargs
    if self.attention is tf.contrib.rnn.AttentionCellWrapper or \
    self.attention is K.rnn_cell.AttentionCell:
      # attn_length: the size of an attention window
      # attn_size: the size of an attention vector. Equal to cell.output_size by default.
      # attn_vec_size: the number of convolutional features calculated
      #     on attention state and a size of the hidden layer built from
      #     base cell state. Equal attn_size to by default.
      attn_length = kwargs.get('attn_length', cell.output_size)
      attn_size = kwargs.get('attn_size', None)
      attn_vec_size = kwargs.get('attn_vec_size', None)
      cell_with_attention = self.attention(cell,
          attn_length=attn_length, attn_size=attn_size,
          attn_vec_size=attn_vec_size)
    else: # seq2seq attention
      num_units = kwargs.get('num_units', X.shape[-1].value)
      attention_mechanism = self.attention(
          num_units=num_units, memory=memory, **kwargs)
      cell_with_attention = tf.contrib.seq2seq.AttentionWrapper(
          cell=cell, attention_mechanism=attention_mechanism,
          attention_layer_size=attention_mechanism._num_units)
    return cell_with_attention

  # ==================== Cells ==================== #
  @property
  def cell(self):
    # ====== first time create Cell ====== #
    if not hasattr(self, '_cell'):
      self._cell = self.__cell_creator()
    # ====== return Cell ====== #
    self._cell._reuse = False if len(self.variables) == 0 else True
    return self._cell

  @property
  def cell_bw(self):
    if not self.bidirectional:
      raise RuntimeError("`cell_bw` is not supported with `bidirectional=False`")
    # ====== first time create Cell ====== #
    if not hasattr(self, '_cell_bw'):
      self._cell_bw = self.__cell_creator()
    # ====== return Cell ====== #
    self._cell_bw._reuse = False if len(self.variables) == 0 else True
    return self._cell_bw

  def _apply(self, X, state=None, memory=None):
    # time_major: The shape format of the `inputs` and `outputs` Tensors.
    #   If true, these `Tensors` must be shaped `[max_time, batch_size, depth]`.
    #   If false, these `Tensors` must be shaped `[batch_size, max_time, depth]`.
    # ====== create attention if necessary ====== #
    cell = self.cell
    if self.bidirectional:
      cell_bw = self.cell_bw
    # create attention cell
    if self.attention:
      if not hasattr(self, "_cell_with_attention"):
        self._cell_with_attention = self.__attention_creator(
            cell, X=X, memory=memory)
        cell = self._cell_with_attention
      # bidirectional attention
      if self.bidirectional:
        if not hasattr(self, "_cell_with_attention_bw"):
          self._cell_with_attention_bw = self.__attention_creator(
              cell_bw, X=X, memory=memory)
        cell_bw = self._cell_with_attention_bw
    # ====== calling rnn_warpper ====== #
    ## Bidirectional
    if self.bidirectional:
      rnn_func = rnn.bidirectional_dynamic_rnn if self.dynamic \
          else rnn.static_bidirectional_rnn
      state_fw, state_bw = None, None
      if isinstance(state, (tuple, list)):
        state_fw = state[0]
        if len(state) > 1:
          state_bw = state[1]
      else:
        state_fw = state
      outputs = rnn_func(cell_fw=cell, cell_bw=cell_bw, inputs=X,
                         initial_state_fw=state_fw,
                         initial_state_bw=state_bw,
                         dtype=X.dtype.base_dtype)
    ## Unidirectional
    else:
      rnn_func = rnn.dynamic_rnn if self.dynamic else rnn.static_rnn
      outputs = rnn_func(cell, inputs=X, initial_state=state,
                         dtype=X.dtype.base_dtype)
    # ====== initialize cell ====== #
    if not self._is_initialized_variables:
      # initialize only once, everytime you call this, the values of
      # variables changed
      K.eval(tf.variables_initializer(self.variables))
      self._is_initialized_variables = True
      _infer_variable_role(self.variables)
    # ====== return ====== #
    if self.bidirectional: # concat outputs
      outputs = (tf.concat(outputs[0], axis=-1), outputs[1])
    if not self.return_states:
      return outputs[0]
    return outputs

  def _transpose(self):
    raise NotImplementedError


class LSTM(RNN):

  def __init__(self, num_units, use_peepholes=False, cell_clip=None,
               num_proj=None, proj_clip=None, forget_bias=1.0,
               activation=None, num_layers=1, bidirectional=False,
               attention=None, attention_kwargs={},
               dynamic=True, return_states=False, name=None):
    super(LSTM, self).__init__(cell_type=tf.contrib.rnn.LSTMCell,
        kwargs={'num_units': num_units, 'use_peepholes': use_peepholes,
                'cell_clip': cell_clip, 'num_proj': num_proj,
                'proj_clip': proj_clip, 'forget_bias': forget_bias,
                'activation': activation},
        num_layers=num_layers, bidirectional=bidirectional,
        attention=attention, attention_kwargs=attention_kwargs,
        dynamic=dynamic, return_states=return_states,
        name=name)


class GRU(RNN):

  def __init__(self, num_units, activation=None,
               attention=None, attention_kwargs={},
               num_layers=1, bidirectional=False, dynamic=True,
               return_states=False, name=None):
    super(GRU, self).__init__(cell_type=tf.contrib.rnn.GRUCell,
        kwargs={'num_units': num_units, 'activation': activation},
        num_layers=num_layers, bidirectional=bidirectional,
        attention=attention, attention_kwargs=attention_kwargs,
        dynamic=dynamic, return_states=return_states,
        name=name)


# ===========================================================================
# DNN
# ===========================================================================
class CudnnRNN(NNOp):

  """CuDNN v5 RNN implementation.

  Parameters
  ----------
  num_units : int
      the number of units within the RNN model.
  W_init:
      initial description for weights
  b_init:
      initial description for bias
  rnn_mode : {'rnn_relu', 'rnn_tanh', 'lstm', 'gru'}
      See cudnn documentation for ``cudnnRNNMode_t``.
  num_layers : int
      the number of layers for the RNN model.
  is_bidirectional : bool (default: False)
      unidirectional: The network operates recurrently from the
                      first input to the last.
      bidirectional: The network operates from first to last then from last
                     to first and concatenates the results at each layer.
  is_training : {None, boolean}
      if None, is_training is conditioned on `odin.backend.is_training()`
  return_states: boolean (defaults: False)
      if True, this Ops returns the [output, hidden_staes, cell_states (lstm)]
      otherwise only return the output
  dropout: float (0.0-1.0)
      whether to enable dropout. With it is 0, dropout is disabled.

  Arguments
  ---------
  X : Tensor
    (batch_size, time_dim, feature_dim)
  h0 : {None, number, Tensor} (default: None)
    if None, all-zeros initial states are used
    (num_layers * num_direction, batch_size, num_units)
  c0 : {None, number, Tensor} (default: None)
    if None, all-zeros initial cell memory are used
    (num_layers * num_direction, batch_size, num_units)
  training : bool (default: None)
    if None, use O.D.I.N training flag,
    otherwise, use given value

  Returns
  -------
  [output, hidden_states, cell_states] for lstm
  [output, hidden_states] for gru and rnn

  output_shape: (batch_size, timesteps,  num_units)
  hidden_shape: (num_layers, batch_size, num_units)
  cell_shape: (num_layers, batch_size,   num_units)

  Note
  ----
  `__call__(training=True)` if you want to get gradient from this NNOp
  """

  def __init__(self, num_units,
          W_init=init_ops.glorot_uniform_initializer(seed=randint()),
          b_init=init_ops.constant_initializer(0.),
          rnn_mode='lstm', num_layers=1,
          skip_input=False, is_bidirectional=False,
          return_states=False, dropout=0., **kwargs):
    super(CudnnRNN, self).__init__(**kwargs)
    # ====== defaults recurrent control ====== #
    self.num_units = int(num_units)
    self.num_layers = int(num_layers)
    self.rnn_mode = str(rnn_mode)
    self.skip_input = bool(skip_input)
    self.is_bidirectional = bool(is_bidirectional)
    self.return_states = bool(return_states)
    self.dropout = dropout

    self.W_init = W_init
    self.b_init = b_init
    if skip_input:
      wprint("`skip_input` is not supported in Tensorflow.")

  # ==================== abstract methods ==================== #
  @property
  def state_shape(self):
    return [self.num_layers * 2 if self.is_bidirectional else 1, None, self.num_units]

  def _transpose(self):
    # flip the input and hidden
    raise NotImplementedError

  def _initialize(self):
    input_shape = self.input_shape_map['X']
    weights, biases = K.init_rnn(
        input_dim=int(input_shape[-1]), hidden_dim=int(self.num_units),
        W_init=self.W_init, b_init=self.b_init,
        num_layers=self.num_layers, num_gates=self.rnn_mode,
        skip_input=self.skip_input,
        is_bidirectional=self.is_bidirectional,
        cudnn_vector=False)
    self._weights_name = [w.name for w in weights]
    self._biases_name = [b.name for b in biases]
    for i in weights + biases:
      self.get_variable_nnop(name=i.name.split('/')[-1].split(':')[0],
                             shape=i.shape.as_list(), initializer=i)

  def _apply(self, X, h0=None, c0=None, training=None):
    if not hasattr(self, '_opaque_params'):
      weights = [K.get_all_variables(full_name=w)[0] for w in self._weights_name]
      biases = [K.get_all_variables(full_name=b)[0] for b in self._biases_name]
      self._opaque_params = K.params_to_cudnn(weights, biases)
    params = self._opaque_params
    outputs = K.cudnn_rnn(X=X, h0=h0, c0=c0,
                          num_units=self.num_units, rnn_mode=self.rnn_mode,
                          num_layers=self.num_layers, parameters=params,
                          skip_input=self.skip_input,
                          is_bidirectional=self.is_bidirectional,
                          dropout=self.dropout,
                          is_training=training)
    if not self.return_states:
      return outputs[0]
    return outputs



from __future__ import print_function, division, absolute_import

import time
import pickle

import numpy as np
import tensorflow as tf
from tensorflow.python.ops import init_ops

from odin.ml.base import BaseEstimator, Evaluable
from odin.backend.role import has_roles, Weight, Bias
from odin import (backend as K, nnet as N, fuel as F, visual as V)
from odin.utils import (is_number, uuid, batching, as_tuple, Progbar,
                        one_hot, wprint, ctext, is_string)

# ===========================================================================
# Helper
# ===========================================================================
def _create_it_func(X, y, batch_size, start, end):
  """ Return a lambda function that create new generator """
  nb_samples = end - start
  create_it = None
  # ====== y is None ====== #
  if y is None:
    if hasattr(X, 'set_batch'):
      create_it = lambda seed: (x for x in X.set_batch(
          batch_size=batch_size,
          start=start, end=end,
          seed=seed))
    elif hasattr(X, '__getitem__'):
      create_it = lambda seed: (X[start:end]
                        for start, end in batching(n=nb_samples,
                                                   batch_size=batch_size,
                                                   start=start, end=end,
                                                   seed=seed))
  # ====== provided y ====== #
  else:
    if hasattr(X, 'set_batch') and hasattr(y, 'set_batch'):
      create_it = lambda seed: ((i, j) for i, j in zip(
          X.set_batch(batch_size=batch_size, start=start, end=end, seed=seed),
          y.set_batch(batch_size=batch_size, start=start, end=end, seed=seed)))
    elif hasattr(X, '__getitem__') and hasattr(y, '__getitem__'):
      create_it = lambda seed: ((X[start:end], y[start:end])
        for start, end in batching(n=nb_samples,
                                   batch_size=batch_size,
                                   start=start, end=end,
                                   seed=seed))
  # ====== exception ====== #
  if create_it is None:
    raise ValueError("`X` and `y` must has attributes 'set_batch' or '__getitem__'")
  return create_it

def _preprocess_xy(x, y, nb_classes):
  if x.ndim > 2:
    x = np.reshape(x, newshape=(x.shape[0], -1))
  if y is not None:
    if y.ndim == 1 and nb_classes > 2:
      y = one_hot(y, nb_classes=nb_classes)
    return x, y
  return x

def _fitting_helper(it, fn, nb_samples, nb_classes, title):
  prog = Progbar(target=nb_samples, print_report=True,
                 print_summary=False, name=title)
  results = None
  start_time = time.time()
  for nb_iter, (x, y) in enumerate(it):
    # ====== preprocessing ====== #
    x, y = _preprocess_xy(x, y, nb_classes)
    # ====== post-processing results ====== #
    if results is None:
      results = list(fn(x, y))
    else:
      for idx, r in enumerate(fn(x, y)):
        results[idx] += r
    # ====== update progress ====== #
    prog.add(x.shape[0])
  duration = time.time() - start_time
  return (nb_iter + 1,
          duration,
          [r if isinstance(r, np.ndarray) else r / (nb_iter + 1)
           for r in results])

# ====== get all optimizer ====== #
_optimizer_list = {}
for name in dir(K.optimizers):
  obj = getattr(K.optimizers, name)
  if isinstance(obj, type) and issubclass(obj, K.optimizers.Optimizer):
    _optimizer_list[name.lower()] = obj
# ===========================================================================
# Logistic regression
# ===========================================================================
class LogisticRegression(BaseEstimator, Evaluable):
  """ LogisticRegression

  Parameters
  ----------
  nb_classes : int
      number of output classes
  l1 : float (scalar)
      weight for L1 regularization
  l2 : float (scalar)
      weight for L2 regularization
  fit_intercept : bool (default: True)
      Specifies if a constant (a.k.a. bias or intercept) should be
      added to the decision function.
  confusion_matrix : bool (default: True)
      print confusion matrix during fitting and evaluating
  tol : float (default: 1e-4)
      Tolerance for stopping criteria
  patience : int (default: 3)
      how many failed improvement before stopping the training
  rollback : bool (default: True)
      allow the model to rollback to the best checkpoint after
      each generalization degradation
  batch_size : int
      batch size
  max_epoch : int
      maximum number of epoch for fitting
  max_iter : int
      maximum number of iteration for fitting
  optimizer : str
      name of optimizer (lower case must match the name in
      `odin.backend.optimizers`)
  learning_rate : float
      learning rate for the optimizer
  class_weight : {float, None, list of scalar}
      weight for each classes during training
  verbose : {bool, int}
      `0` or `False`, totally turn off all logging during fitting
      `1` or `True`, turn on the summary for fitting
       >= `2`, turn on progress of each iteration and summary
  path : str
      checkpoint path for backing the model after every
      improvement
  name : str
      specific name for tensorflow Op
  """

  def __init__(self, nb_classes, l1=0., l2=0.,
               fit_intercept=True, confusion_matrix=True,
               tol=1e-4, patience=3, rollback=True,
               batch_size=1024, max_epoch=100, max_iter=None,
               optimizer='adadelta', learning_rate=1.0, class_weight=None,
               dtype='float32', seed=5218,
               verbose=False, path=None, name=None):
    super(LogisticRegression, self).__init__()
    # ====== basic dimensions ====== #
    if isinstance(nb_classes, (tuple, list, np.ndarray)):
      self._labels = tuple([str(i) for i in nb_classes])
      self._nb_classes = len(nb_classes)
    elif is_number(nb_classes):
      self._labels = tuple([str(i) for i in range(nb_classes)])
      self._nb_classes = int(nb_classes)
    self._feat_dim = None
    self._dtype = np.dtype(dtype)
    # ====== preprocessing class weight ====== #
    if class_weight is None:
      class_weight = np.ones(shape=(self.nb_classes,),
                             dtype=self.dtype)
    elif is_number(class_weight):
      class_weight = np.zeros(shape=(self.nb_classes,),
                              dtype=self.dtype) + class_weight
    self._class_weight = class_weight
    # ====== flags ====== #
    self.l1 = float(l1)
    self.l2 = float(l2)
    self.fit_intercept = bool(fit_intercept)
    self.confusion_matrix = bool(confusion_matrix)
    # ====== internal states ====== #
    self._is_fitted = False
    # ====== others ====== #
    if name is None:
      name = uuid(length=8)
      self._name = 'LogisticRegression_%s' % name
    else:
      self._name = str(name)
    self._path = path
    # ====== training ====== #
    self.batch_size = int(batch_size)
    self.max_epoch = max_epoch
    self.max_iter = max_iter
    if not is_string(optimizer):
      raise ValueError("`optimizer` must be one of the following")
    optimizer = optimizer.lower()
    if optimizer not in _optimizer_list:
      raise ValueError("`optimizer` must be one of the following: %s" %
        str(list(_optimizer_list.keys())))
    self._optimizer = _optimizer_list[optimizer.lower()](lr=float(learning_rate))
    self._optimizer_name = optimizer
    self._optimizer_lr = learning_rate
    # ====== stop training ====== #
    self.tol = float(tol)
    self.patience = int(patience)
    self.rollback = bool(rollback)
    # ====== others ====== #
    self._train_history = []
    self._valid_history = []
    self._rand_state = np.random.RandomState(seed=int(seed))
    self.verbose = int(verbose)

  # ==================== pickling ==================== #
  def __getstate__(self):
    return (self._labels, self._nb_classes, self._feat_dim, self._dtype,
      self._class_weight, self.l1, self.l2, self.fit_intercept,
      self.confusion_matrix, self._is_fitted, self._name, self._path,
      self.batch_size, self.max_epoch, self.max_iter,
      self.tol, self.patience, self.rollback,
      self._train_history, self._valid_history, self._rand_state, self.verbose,
      self._optimizer_name, self._optimizer_lr, self.parameters)

  def __setstate__(self, states):
    (self._labels, self._nb_classes, self._feat_dim, self._dtype,
     self._class_weight, self.l1, self.l2, self.fit_intercept,
     self.confusion_matrix, self._is_fitted, self._name, self._path,
     self.batch_size, self.max_epoch, self.max_iter,
     self.tol, self.patience, self.rollback,
     self._train_history, self._valid_history, self._rand_state, self.verbose,
     self._optimizer_name, self._optimizer_lr, parameters) = states
    self._optimizer = _optimizer_list[self._optimizer_name](
        lr=float(self._optimizer_lr))
    self._initialize(X=np.empty(shape=(1, self.feat_dim),
                                dtype=self.dtype))
    self.set_parameters(*parameters)

  # ==================== properties ==================== #
  @property
  def name(self):
    return self._name

  @property
  def path(self):
    return self._path

  @property
  def labels(self):
    return tuple(self._labels)

  @property
  def nb_classes(self):
    return self._nb_classes

  @property
  def feat_dim(self):
    return self._feat_dim

  @property
  def is_fitted(self):
    return self._is_fitted

  @property
  def dtype(self):
    return self._dtype

  @property
  def placeholders(self):
    return (self._X, self._y)

  @property
  def parameters(self):
    """ Return list of parameters
    [W, b] if fit_intercept=True
    [W] otherwise
    """
    if self.fit_intercept:
      return [K.get_value(self._model.get('W')),
              K.get_value(self._model.get('b'))]
    else:
      return [K.get_value(self._model.get('W'))]

  def set_parameters(self, W=None, b=None):
    if W is not None:
      K.set_value(self._model.get('W'), W)
    if self.fit_intercept and b is not None:
      K.set_value(self._model.get('b'), b)
    return self

  # ==================== sklearn methods ==================== #
  def _initialize(self, X):
    # ====== check inputs dimensions ====== #
    if not hasattr(X, 'shape'):
      raise ValueError("`X` must have `shape` attribute.")
    feat_dim = np.prod(X.shape[1:])
    if self._feat_dim is None:
      self._feat_dim = feat_dim
    # validate input dimension
    if feat_dim != self._feat_dim:
      raise RuntimeError("Feature dimension mismatch %d and %d" %
                         (feat_dim, self.feat_dim))
    # check if tensorflow op initalized
    if hasattr(self, '_f_train'):
      return
    # ====== binary or multi-classes ====== #
    if self.nb_classes == 2:
      out_shape = (None,)
      fn_activation = tf.nn.sigmoid
      fn_loss = tf.losses.sigmoid_cross_entropy
      fn_acc = K.metrics.binary_accuracy
    else:
      out_shape = (None, self.nb_classes)
      fn_activation = tf.nn.softmax
      fn_loss = tf.losses.softmax_cross_entropy
      fn_acc = K.metrics.categorical_accuracy
    # ====== create model ====== #
    with tf.name_scope(self.name, 'logistic_regression'):
      # inputs
      self._X = K.placeholder(shape=(None, self.feat_dim),
                              dtype=self.dtype,
                              name='%s_input' % self.name)
      self._y = K.placeholder(shape=out_shape,
                              dtype=self.dtype,
                              name='%s_output' % self.name)
      # check the bias
      if is_number(self.fit_intercept):
        b_init = float(self.fit_intercept)
      elif self.fit_intercept is False or \
      self.fit_intercept is None:
        b_init = None
      else:
        b_init = self.fit_intercept
      # create the model and initialize
      with K.variable_dtype(dtype=self.dtype):
        self._model = N.Dense(num_units=self.nb_classes,
                          W_init=init_ops.glorot_uniform_initializer(seed=self._rand_state.randint()),
                          b_init=b_init,
                          activation=K.linear)
        y_logits = self._model(self._X)
      y_prob = fn_activation(y_logits)
      # applying class weights
      class_weights = tf.constant(value=self._class_weight,
                                  dtype=self.dtype,
                                  name="class_weights")
      weights = tf.gather(class_weights,
                          tf.cast(self._y, 'int32') if self.nb_classes == 2 else
                          tf.argmax(self._y, axis=-1))
      # optimizer
      params = [v for v in self._model.variables
                if has_roles(v, Weight) or has_roles(v, Bias)]
      losses = fn_loss(self._y, y_logits, weights=weights)
      l1_norm = tf.norm(self._model.get('W'), ord=1) if self.l1 > 0. else 0
      l2_norm = tf.norm(self._model.get('W'), ord=2) if self.l2 > 0. else 0
      losses = losses + self.l1 * l1_norm + self.l2 * l2_norm
      acc = fn_acc(self._y, y_prob)
      updates = self._optimizer.get_updates(losses, params)
      # create function
      if self.confusion_matrix:
        cm = K.metrics.confusion_matrix(y_true=self._y, y_pred=y_prob,
                                        labels=self.nb_classes)
      metrics = [losses, acc, cm] if self.confusion_matrix else [losses, acc]
      self._f_train = K.function(inputs=(self._X, self._y),
                                 outputs=metrics,
                                 updates=updates,
                                 training=True)
      self._f_score = K.function(inputs=(self._X, self._y),
                                 outputs=metrics,
                                 training=False)
      self._f_pred_prob = K.function(inputs=self._X,
                                     outputs=y_prob,
                                     training=False)
      self._f_pred_logit = K.function(inputs=self._X,
                                      outputs=y_logits,
                                      training=False)
    return self

  def fit(self, X, y=None, cv=None):
    self._initialize(X)
    if not hasattr(X, 'shape') or not hasattr(X, '__iter__') or \
    not hasattr(X, '__len__'):
      raise ValueError("`X` must has 'shape', '__len__' and '__iter__' attributes")
    nb_train_samples = len(X)
    # convert to odin.fuel.Data if possible
    if isinstance(X, (np.ndarray, list, tuple)):
      X = F.as_data(X)
    if isinstance(y, (np.ndarray, list, tuple)):
      y = F.as_data(y)
    start_tr = 0
    end_tr = nb_train_samples
    # ====== check if cross validating ====== #
    create_it_cv = None
    if is_number(cv):
      cv = int(float(cv) * nb_train_samples) if cv < 1. else int(cv)
      end_tr = nb_train_samples - cv
      start_cv = end_tr
      end_cv = nb_train_samples
      nb_cv_samples = end_cv - start_cv
      create_it_cv = _create_it_func(X=X, y=y, batch_size=self.batch_size,
                                     start=start_cv, end=end_cv)
    elif isinstance(cv, (tuple, list)):
      X_cv, y_cv = cv
      nb_cv_samples = X_cv.shape[0]
      create_it_cv = _create_it_func(X=X_cv, y=y_cv, batch_size=self.batch_size,
                                     start=0, end=X_cv.shape[0])
    elif hasattr(cv, 'set_batch'):
      nb_cv_samples = cv.shape[0]
      create_it_cv = _create_it_func(X=cv, y=None, batch_size=self.batch_size,
                                     start=0, end=cv.shape[0])
    elif cv is not None:
      raise ValueError('`cv` can be float (0-1), tuple or list of X and y, '
                       'any object that have "shape" and "__iter__" attributes, '
                       'or None')
    # ====== preprocessing ====== #
    create_it = _create_it_func(X=X, y=y, batch_size=self.batch_size,
                                start=start_tr, end=end_tr)
    # ====== prepare ====== #
    curr_niter = sum(epoch[0] for epoch in self._train_history)
    curr_nepoch = len(self._train_history)
    curr_patience = int(self.patience)
    last_losses = None
    last_checkpoint = None
    best_epoch = None
    is_converged = False
    # ====== fitting ====== #
    while not is_converged:
      curr_nepoch += 1
      seed = self._rand_state.randint(0, 10e8)
      # ====== training ====== #
      nb_iter, duration, results = _fitting_helper(create_it(seed),
                                                   fn=self._f_train,
                                                   nb_samples=nb_train_samples,
                                                   nb_classes=self.nb_classes,
                                                   title='Epoch %d' % curr_nepoch)
      curr_niter += nb_iter
      self._train_history.append(
          (nb_train_samples, nb_iter, duration, results))
      # ====== cross validation ====== #
      if create_it_cv is not None:
        nb_iter, duration_valid, results = _fitting_helper(create_it_cv(seed),
                                                     fn=self._f_score,
                                                     nb_samples=nb_cv_samples,
                                                     nb_classes=self.nb_classes,
                                                     title="Validating")
        self._valid_history.append(
            (nb_train_samples, nb_iter, duration_valid, results))
        duration += duration_valid
      # ====== print log ====== #
      if self.verbose >= 2:
        print(ctext('#epoch:', 'cyan') + str(curr_nepoch),
              ctext('#iter:', 'cyan') + str(curr_niter),
              ctext("Loss:", 'yellow') + '%.5f' % results[0],
              ctext("Acc:", 'yellow') + '%.3f' % results[1],
              ctext("%.2f(s)" % duration, 'magenta'))
        if self.confusion_matrix and (curr_nepoch - 1) % 8 == 0:
          print(V.print_confusion(results[-1], labels=self.labels))
      # ====== early stopping ====== #
      losses = results[0]
      if last_checkpoint is None: # first check point
        last_checkpoint = self.parameters
      if last_losses is not None:
        # degraded, smaller is better
        if last_losses - losses <= self.tol:
          curr_patience -= 1
          if self.rollback:
            if self.verbose >= 2:
              wprint('[LogisticRegression] Rollback to the best checkpoint '
                     'at epoch:%s patience:%s' %
                     (ctext(best_epoch, 'cyan'),
                      ctext(curr_patience, 'cyan')))
            self.set_parameters(*last_checkpoint)
        # save best checkpoint
        else:
          last_checkpoint = self.parameters
          best_epoch = curr_nepoch
          if self._path is not None:
            with open(self._path, 'wb') as f:
              pickle.dump(self, f)
      last_losses = losses
      if curr_patience <= 0:
        is_converged = True
      # end the training
      if self.max_iter is not None and \
      curr_niter >= self.max_iter:
        break
      if self.max_epoch is not None and \
      curr_nepoch >= self.max_epoch:
        break
    # ====== print summary plot ====== #
    if self.verbose >= 1:
      train_losses = [epoch[-1][0] for epoch in self._train_history]
      print(V.print_bar(train_losses, height=12,
                        bincount=min(20, len(train_losses)),
                        title='Training Losses'))
      if create_it_cv is not None:
        valid_losses = [epoch[-1][0] for epoch in self._valid_history]
        print(V.print_bar(valid_losses, height=12,
                          bincount=min(20, len(train_losses)),
                          title='Validation Losses'))
      if self.confusion_matrix:
        print(ctext("======== Training Confusion Matrix ========", 'cyan'))
        print(V.print_confusion(arr=self._train_history[-1][-1][-1],
                                labels=self.labels))
        if create_it_cv is not None:
          print(ctext("======== Validation Confusion Matrix ========", 'cyan'))
          print(V.print_confusion(arr=self._valid_history[-1][-1][-1],
                                  labels=self.labels))
    # ====== reset to best points ====== #
    self.set_parameters(*last_checkpoint)
    self._is_fitted = True
    if self._path is not None:
      with open(self._path, 'wb') as f:
        pickle.dump(self, f)

  def predict(self, X):
    """Predict class labels for samples in X.

    Parameters
    ----------
    X : {array-like}, shape = [nb_samples, feat_dim]
        Samples.

    Returns
    -------
    C : array, shape = [nb_samples]
        Predicted class label per sample.
    """
    return np.argmax(self.predict_proba(X), axis=-1)

  def _predict(self, X, f_pred):
    if not self.is_fitted:
      raise RuntimeError("LogisticRegression hasn't been initialized or "
                         "fitted.")
    if hasattr(X, 'set_batch'):
      it = iter(X.set_batch(batch_size=self.batch_size, seed=None))
    elif hasattr(X, '__getitem__'):
      it = (X[start:end]
            for start, end in batching(batch_size=self.batch_size,
                                       n=X.shape[0]))
    else:
      raise ValueError("`X` must has attributes 'set_batch' or '__getitem__'")
    # ====== make prediction ====== #
    y = []
    prog = Progbar(target=X.shape[0], print_report=True,
                   print_summary=False, name="Predicting")
    for x in it:
      x = _preprocess_xy(x, y=None, nb_classes=self.nb_classes)
      y.append(f_pred(x))
      prog.add(x.shape[0])
    return np.concatenate(y, axis=0)

  def predict_logits(self, X):
    """Logits values estimates.
    The returned estimates for all classes are ordered by the
    label of classes.
    """
    return self._predict(X, f_pred=self._f_pred_logit)

  def predict_proba(self, X):
    """Probability estimates.
    The returned estimates for all classes are ordered by the
    label of classes.

    Parameters
    ----------
    X : array-like, shape = [nb_samples, feat_dim]

    Returns
    -------
    y : array-like, shape = [nb_samples, nb_classes]
        Returns the probability of the sample for each class in the model,
        where classes are ordered as they are in ``self.classes_``.
    """
    return self._predict(X, f_pred=self._f_pred_prob)

  def predict_log_proba(self, X):
    """Log of probability estimates.
    The returned estimates for all classes are ordered by the
    label of classes.
    Parameters
    ----------
    X : array-like, shape = [n_samples, n_features]
    Returns
    -------
    T : array-like, shape = [n_samples, n_classes]
        Returns the log-probability of the sample for each class in the
        model, where classes are ordered as they are in ``self.classes_``.
    """
    return np.log(self.predict_proba(X))



from __future__ import print_function, division, absolute_import

import inspect
from six import add_metaclass
from collections import Mapping
from abc import ABCMeta, abstractmethod

import numpy as np

from odin.utils import (flatten_list, is_string, as_tuple,
                        ctext, is_number, is_primitives)

# ===========================================================================
# Recipes
# ===========================================================================
@add_metaclass(ABCMeta)
class FeederRecipe(object):
  """ All method of this function a called in following order
  preprocess_indices(indices): return new_indices
  init(ntasks, batch_size, seed): right before create the iter

  [multi-process] process(*x): x->(name, data)
                               return (if iterator, it will be iterated to
                                       get a list of results)
  [multi-process] group(x): x->(object from group(x))
                            return iterator

  Note
  ----
  This class should not store big amount of data, or the data
  will be replicated to all processes
  """

  def __init__(self):
    super(FeederRecipe, self).__init__()
    self._nb_desc = 0

  # ==================== basic properties ==================== #
  def set_feeder_info(self, nb_desc=None):
    if nb_desc is not None:
      self._nb_desc = int(nb_desc)
    return self

  @property
  def nb_desc(self):
    return self._nb_desc

  # ==================== abstract ==================== #
  def shape_transform(self, shapes):
    """
    Parameters
    ----------
    shapes: list of [(shape0, indices0), (shape1, indices1), ...]
        list of data shape tuple and indices, the indices is list
        of tuple (name, length)

    Return
    ------
    new shape that transformed by this Recipe
    new indices
    """
    return shapes

  @abstractmethod
  def process(self, name, X):
    """
    Parameters
    ----------
    name: string
        the name of file in indices
    X: list of data
        list of all features given in IndexedData(s)
    """
    raise NotImplementedError

  def __str__(self):
    # ====== get all attrs ====== #
    all_attrs = dir(self)
    print_attrs = {}
    for name in all_attrs:
      if '_' != name[0] and (len(name) >= 2 and '__' != name[:2]) and\
      name not in ('nb_desc'):
        attr = getattr(self, name)
        if name == 'data_idx':
          print_attrs[name] = str(attr)
        elif isinstance(attr, slice):
          print_attrs[name] = str(attr)
        elif inspect.isfunction(attr):
          print_attrs[name] = "(f)" + attr.__name__
        elif isinstance(attr, np.ndarray):
          print_attrs[name] = ("(%s)" % str(attr.dtype)) + \
              str(attr.shape)
        elif isinstance(attr, (tuple, list)):
          print_attrs[name] = "(list)" + str(len(attr))
        elif isinstance(attr, Mapping):
          print_attrs[name] = "(map)" + str(len(attr))
        elif is_primitives(attr):
          print_attrs[name] = str(attr)
    print_attrs = sorted(print_attrs.items(), key=lambda x: x[0])
    print_attrs = [('#desc', self.nb_desc)] + print_attrs
    print_attrs = ' '.join(["%s:%s" % (ctext(key, 'yellow'), val)
                            for key, val in print_attrs])
    # ====== format the output ====== #
    s = '<%s %s>' % (ctext(self.__class__.__name__, 'cyan'), print_attrs)
    return s

  def __repr__(self):
    return self.__str__()


class RecipeList(FeederRecipe):

  def __init__(self, *recipes):
    super(RecipeList, self).__init__()
    self._recipes = recipes

  # ==================== List methods ==================== #
  def __iter__(self):
    return self._recipes.__iter__()

  def __len__(self):
    return len(self._recipes)

  # ==================== Override ==================== #
  def set_recipes(self, *recipes):
    # filter out None value
    recipes = flatten_list(as_tuple(recipes))
    recipes = [rcp for rcp in recipes
               if rcp is not None and isinstance(rcp, FeederRecipe)]
    # ====== set the new recipes ====== #
    if len(recipes) > 0:
      self._recipes = recipes
      for rcp in self._recipes:
        rcp.set_feeder_info(self.nb_desc)
    return self

  def set_feeder_info(self, nb_desc=None):
    super(RecipeList, self).set_feeder_info(nb_desc)
    for rcp in self._recipes:
      rcp.set_feeder_info(nb_desc)
    return self

  def process(self, name, X, **kwargs):
    for i, f in enumerate(self._recipes):
      # return iterator (iterate over all of them)
      args = f.process(name, X)
      # break the chain if one of the recipes get error,
      # and return None
      if args is None:
        return None
      if not isinstance(args, (tuple, list)) or \
      len(args) != 2 or \
      not is_string(args[0]) or \
      not isinstance(args[1], (tuple, list)):
        raise ValueError("The returned from `process` must be tuple or "
            "list, of length 2 which contains (name, [x1, x2, x3,...])."
            "`name` must string type, and [x1, x2, ...] is tuple or list.")
      name, X = args
    return name, X

  def shape_transform(self, shapes):
    """
    Parameters
    ----------
    shapes: list of [(shape0, indices0), (shape1, indices1), ...]
        list of data shape tuple and indices, the indices is list
        of tuple (name, length)

    Return
    ------
    new shape that transformed by this Recipe
    new indices
    """
    for i in self._recipes:
      shapes = i.shape_transform(shapes)
      # ====== check returned ====== #
      if not all((isinstance(shp, (tuple, list)) and
                  all(is_number(s) for s in shp) and
                  is_string(ids[0][0]) and is_number(ids[0][1]))
                 for shp, ids in shapes):
        raise RuntimeError("Returned `shapes` must be the list of pair "
                           "`(shape, indices)`, where `indices` is the "
                           "list of (name, length(int)).")
    return shapes



from __future__ import print_function, division, absolute_import
import os
import math
import types
import inspect
import warnings
from collections import Counter, Mapping
from six.moves import cPickle
from six.moves import zip, zip_longest, range

import numpy as np

from odin.utils import (one_hot, is_string, ctext, is_number,
                        is_primitives, as_tuple, flatten_list,
                        is_pickleable)
from odin.fuel.utils import MmapDict
from odin.fuel.recipe_base import FeederRecipe
from odin.fuel.recipe_shape import *
from odin.fuel.recipe_norm import *


# ===========================================================================
# Loader
# ===========================================================================
class TransLoader(FeederRecipe):
  """ Load transcription from dictionary using name specifed in
  indices

  map_func:
      inputs: name, X
      outputs: name, X, loaded_transcription
  reduce_func:
      inputs: same
      outputs: same

  Parameters
  ----------
  transcription: dict
      if path to a file is specified, the file must specified
      <name> -> [frame1, frame2, ...]
      if list is given, the list must contain the same information
      if dictionary is given, the dict must repesent the same mapping
      above

  """

  def __init__(self, transcription, dtype, delimiter=' ', label_dict=None,
               ignore_not_found=True):
    super(TransLoader, self).__init__()
    raise NotImplementedError
    self.ignore_not_found = ignore_not_found
    # ====== transcription ====== #
    share_dict = None
    if isinstance(transcription, (list, tuple)):
      share_dict = {i: j for i, j in transcription}
    elif isinstance(transcription, Mapping):
      share_dict = transcription
    elif isinstance(transcription, str) and os.path.isfile(transcription):
      share_dict = MmapDict(transcription)
    else:
      raise Exception('Cannot understand given transcription information.')
    self._transcription = share_dict
    # ====== datatype and delimiter ====== #
    # NO 64bit data type
    self.dtype = str(np.dtype(dtype)).replace('64', '32')
    self.delimiter = delimiter
    # ====== label dict if available ====== #
    if label_dict is None:
      label_func = lambda x: x
    elif isinstance(label_dict, Mapping):
      label_func = lambda x: label_dict[x]
    elif hasattr(label_dict, '__call__'):
      label_func = label_dict
    else:
      raise ValueError('label_dict must be a dictionary, function or None.')
    self.label_dict = label_func

  def process(self, name, X):
    if name not in self._transcription and self.ignore_not_found:
      return None
    trans = self._transcription[name]
    # ====== parse string using delimiter ====== #
    if isinstance(trans, str):
      trans = [self.label_dict(i)
               for i in trans.split(self.delimiter)
               if len(i) > 0]
    else:
      trans = [self.label_dict(i) for i in trans]
    trans = np.asarray(trans, dtype=self.dtype)
    # append to trans list
    X.append(trans)
    return name, X

  def shape_transform(self, shapes):
    pass


# ===========================================================================
# Basic recipes
# ===========================================================================
class Filter(FeederRecipe):

  """
  Parameters
  ----------
  filter_func: function(name, X)
      return True if the given data is accepted for further processing
      otherwise False

  """

  def __init__(self, filter_func):
    super(Filter, self).__init__()
    if not hasattr(filter_func, '__call__'):
      raise ValueError('"filter_func" must be call-able.')
    if not is_pickleable(filter_func):
      raise ValueError('"filter_func" must be pickle-able.')
    self.filter_func = filter_func

  def process(self, name, X):
    if self.filter_func(name, X):
      return name, X
    return None


# ===========================================================================
# Label processing
# ===========================================================================
class LabelOneHot(FeederRecipe):

  def __init__(self, nb_classes, data_idx=()):
    super(LabelOneHot, self).__init__()
    self.nb_classes = int(nb_classes)
    self.data_idx = data_idx

  def process(self, name, X):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(X),
                              return_tuple=True)
    X_new = []
    for idx, x in enumerate(X):
      # transform into one-label y
      if idx in data_idx:
        x = np.array(x, dtype='int32')
        x = one_hot(x, nb_classes=self.nb_classes)
      X_new.append(x)
    return name, X_new

  def shape_transform(self, shapes):
    data_idx = axis_normalize(axis=self.data_idx,
                              ndim=len(shapes),
                              return_tuple=True)
    return [((shp[0], self.nb_classes), ids) if idx in data_idx else
            (shp, ids)
            for idx, (shp, ids) in enumerate(shapes)]


class Name2Label(FeederRecipe):
  """ This function convert the name (in indices) to transcription
  for given data

  Parameters
  ----------
  converter_func: callbale (1 input arguments)
      for example, lambda name: 1 if 'true' in name else 0
      the return label then is duplicated for all data points in 1 file.
      (e.g. X.shape = (1208, 13), then, transcription=[ret] * 1208)
  ref_idx: int
      the new label will be duplicated based on the length of data
      at given idx

  Example
  -------
  >>> cluster_idx = ['spa-car', 'por-brz', 'spa-lac', 'spa-eur']
  >>> feeder = F.Feeder(ds['mfcc'], ds.path, ncpu=1, buffer_size=12)
  >>> feeder.set_batch(256, seed=None, shuffle_level=0)
  >>> feeder.set_recipes([
  >>>     F.recipes.NameToTranscription(lambda x: cluster_idx.index(x),
  >>>     F.recipes.CreateBatch()
  >>> ])

  """

  def __init__(self, converter_func, dtype=None, ref_idx=0):
    super(Name2Label, self).__init__()
    if inspect.isfunction(converter_func):
      converter_func = converter_func
    if not hasattr(converter_func, '__call__'):
      raise ValueError('"converter_func" must be call-able.')
    self.converter_func = converter_func
    self.ref_idx = int(ref_idx)
    self.dtype = dtype

  def process(self, name, X):
    # X: is a list of ndarray
    ref_idx = axis_normalize(axis=self.ref_idx, ndim=len(X),
                             return_tuple=False)
    y = self.converter_func(name)
    y = np.full(shape=(X[ref_idx].shape[0],),
                fill_value=y,
                dtype=X[ref_idx].dtype if self.dtype is None else self.dtype)
    X.append(y)
    return name, X

  def shape_transform(self, shapes):
    ref_idx = axis_normalize(axis=self.ref_idx, ndim=len(shapes),
                             return_tuple=False)
    ref_shp, ref_ids = shapes[ref_idx]
    ids = list(ref_ids)
    n = int(ref_shp[0])
    shapes.append(((n,), ids))
    return shapes



from odin import nnet as N
from odin import preprocessing as pp
from utils import WAV_FILES

def bnf_sad(debugging):
  bnf_network = N.models.BNF_2048_MFCC40()
  recipe = pp.make_pipeline(steps=[
      pp.speech.AudioReader(sr=16000, sr_new=8000,
                            best_resample=True, remove_dc=True),
      pp.speech.PreEmphasis(coeff=0.97),
      pp.base.Converter(converter=WAV_FILES,
                        input_name='path', output_name='name'),
      # ====== STFT ====== #
      pp.speech.STFTExtractor(frame_length=0.025, step_length=0.010,
                              window='hamm', n_fft=512),
      # ====== SAD ====== #
      pp.base.RenameFeatures(input_name='stft_energy', output_name='energy'),
      pp.speech.SADgmm(nb_mixture=3, smooth_window=3,
                       input_name='energy', output_name='sad'),
      # ====== spectrogram ====== #
      pp.speech.PowerSpecExtractor(power=2.0, output_name='spec'),
      pp.speech.MelsSpecExtractor(n_mels=24, fmin=100, fmax=4000,
                                  input_name='spec', output_name='mspec'),
      pp.speech.MFCCsExtractor(n_ceps=20, remove_first_coef=True,
                               input_name='mspec', output_name='mfcc'),
      pp.base.DeltaExtractor(input_name='mfcc', order=(0, 1, 2)),
      pp.speech.ApplyingSAD(input_name=('mspec', 'mfcc')),
      # ====== BNF ====== #
      pp.speech.MelsSpecExtractor(n_mels=40, fmin=100, fmax=4000,
                                  input_name='spec', output_name='mspec_bnf'),
      pp.speech.MFCCsExtractor(n_ceps=40, remove_first_coef=False,
                               input_name='mspec_bnf', output_name='mfcc_bnf'),
      pp.base.AsType(dtype='float32', input_name='mfcc_bnf'),
      pp.speech.BNFExtractor(input_name='mfcc_bnf', output_name='bnf',
                             sad_name='sad',
                             network=bnf_network,
                             remove_non_speech=True,
                             stack_context=10, pre_mvn=True,
                             batch_size=5218),
      # ====== normalization ====== #
      pp.speech.AcousticNorm(input_name=('mspec', 'bnf', 'mfcc'),
                             mean_var_norm=True,
                             windowed_mean_var_norm=True,
                             win_length=301),
      # ====== cleaning ====== #
      pp.base.DeleteFeatures(input_name=('stft', 'raw', 'energy',
                                         'sad', 'sad_threshold',
                                         'spec', 'mspec_bnf', 'mfcc_bnf')),
      pp.base.AsType(dtype='float16')
  ], debug=debugging)
  return recipe

def bnf_all(debugging):
  bnf_network = N.models.BNF_2048_MFCC40()
  recipe = pp.make_pipeline(steps=[
      pp.speech.AudioReader(sr=16000, sr_new=8000,
                            best_resample=True, remove_dc=True),
      pp.speech.PreEmphasis(coeff=0.97),
      pp.base.Converter(converter=WAV_FILES,
                        input_name='path', output_name='name'),
      # ====== STFT ====== #
      pp.speech.STFTExtractor(frame_length=0.025, step_length=0.010,
                              window='hamm', n_fft=512),
      # ====== SAD ====== #
      pp.base.RenameFeatures(input_name='stft_energy', output_name='energy'),
      pp.speech.SADgmm(nb_mixture=3, smooth_window=3,
                       input_name='energy', output_name='sad'),
      # ====== spectrogram ====== #
      pp.speech.PowerSpecExtractor(power=2.0, output_name='spec'),
      pp.speech.MelsSpecExtractor(n_mels=24, fmin=100, fmax=4000,
                                  input_name='spec', output_name='mspec'),
      pp.speech.MFCCsExtractor(n_ceps=20, remove_first_coef=True,
                               input_name='mspec', output_name='mfcc'),
      pp.base.DeltaExtractor(input_name='mfcc', order=(0, 1, 2)),
      # ====== BNF ====== #
      pp.speech.MelsSpecExtractor(n_mels=40, fmin=100, fmax=4000,
                                  input_name='spec', output_name='mspec_bnf'),
      pp.speech.MFCCsExtractor(n_ceps=40, remove_first_coef=False,
                               input_name='mspec_bnf', output_name='mfcc_bnf'),
      pp.base.AsType(dtype='float32', input_name='mfcc_bnf'),
      pp.speech.BNFExtractor(input_name='mfcc_bnf', output_name='bnf',
                             sad_name='sad',
                             network=bnf_network,
                             remove_non_speech=False,
                             stack_context=10, pre_mvn=True,
                             batch_size=5218),
      # ====== normalization ====== #
      pp.speech.AcousticNorm(input_name=('mspec', 'bnf', 'mfcc'),
                             mean_var_norm=True,
                             windowed_mean_var_norm=True,
                             win_length=301),
      # ====== cleaning ====== #
      pp.base.DeleteFeatures(input_name=('stft', 'raw', 'energy',
                                         'sad', 'sad_threshold',
                                         'spec', 'mspec_bnf', 'mfcc_bnf')),
      pp.base.AsType(dtype='float16')
  ], debug=debugging)
  return recipe

def mfcc_sad(debugging):
  recipe = pp.make_pipeline(steps=[
      pp.speech.AudioReader(sr=16000, sr_new=8000),
      pp.speech.PreEmphasis(coeff=0.97),
      pp.base.Converter(converter=WAV_FILES,
                        input_name='path', output_name='name'),
      # ====== STFT ====== #
      pp.speech.STFTExtractor(frame_length=0.025, step_length=0.01,
                              n_fft=512, energy=False),
      # ====== spectrogram ====== #
      pp.speech.PowerSpecExtractor(power=2.0, output_name='spec'),
      pp.speech.MelsSpecExtractor(n_mels=24, fmin=20, fmax=3700,
                                  output_name='mspec'),
      pp.speech.MFCCsExtractor(n_ceps=24, remove_first_coef=True,
                               first_coef_energy=True,
                               output_name='mfcc'),
      # ====== SAD ====== #
      pp.speech.SADthreshold(energy_threshold=0.5, smooth_window=5,
                             output_name='sad'),
      pp.speech.SADgmm(nb_mixture=3, smooth_window=3,
                       input_name='energy', output_name='sad'),
      pp.speech.ApplyingSAD(input_name=('mspec',)),
      pp.speech.AcousticNorm(input_name=('mspec',), mean_var_norm=True,
                             windowed_mean_var_norm=True, win_length=121),
      # ====== cleaning ====== #
      pp.base.DeleteFeatures(input_name=('stft', 'raw', 'spec', 'sad',
                                         'sad_threshold', 'energy')),
      pp.base.AsType(dtype='float16')
  ], debug=debugging)
  return recipe



# Address: {b6ba96161f8b621ce7d0424c6502c38e2f089fbc}
from __future__ import division, absolute_import, print_function

import os
import re
import shutil
import pickle
from itertools import chain
from collections import defaultdict, OrderedDict
from six.moves import range, zip, cPickle

import numpy as np

from odin.autoconfig import get_rng
from odin.training.callbacks import *
from odin.fuel import Dataset, as_data
from odin import fuel, backend as K, nnet as N
from odin.utils import (struct, as_tuple, is_number, Progbar,
                        add_notification, array_size, ctext)

# ===========================================================================
# Helper
# ===========================================================================
def __format_string(nb_of_float):
  x = ["{:.4f}"] * int(nb_of_float)
  return ";".join(x)

def _plot_each_epoch(name, results, task_type):
  """ results list of each epoch results
  [(epoch1_r1, epoch1_r2, ...),
   (epoch2_r1, epoch2_r2, ...), ...]
  """
  from matplotlib import pyplot as plt
  nb_epoch = len(results)
  ncol = 3; nrow = int(np.ceil(nb_epoch / ncol))
  _ = list(chain(*results))
  max_ = np.max(_); min_ = np.min(_)
  # ====== plot an overall view of all epoch ====== #
  plt.figure()
  line = plt.plot(range(1, len(results) + 1),
           [np.mean(epoch) for epoch in results])[0]
  plt.setp(line, linewidth=2, color='r')
  plt.ylim([min_, max_])
  plt.xlabel("#Epoch")
  plt.ylabel(name)
  plt.suptitle(task_type + name, fontsize=20)
  # ====== plot each epoch ====== #
  plt.figure()
  for i, x in enumerate(results):
    ax = plt.subplot(nrow, ncol, i + 1)
    ax.plot(x); ax.set_ylim([min_, max_])
    ax.tick_params(labelsize=8)
    plt.xlabel("[Epoch%d]Iteration" % (i + 1), fontsize=8)
  plt.tight_layout()

# ===========================================================================
# Tasks
# ===========================================================================
class Task(object):
  """
  Parameters
  ----------
  func: call-able
      function will be executed for each iteration
  data: single or list of odin.fuel.Data, numpy.ndarray
      iterate over all these data and execute function on
      the data.
  epoch: int
      how many epoch will be repeated
  p: float (0.0 - 1.0)
      probability the `func` will be execute for each iteration
  batch_size: int (> 0)
      number of samples for each iteration
  seed: int
      random seed for shuffling the data
  shuffle_level: int (0, 1, 2)
      if 0, shuffle the file lists
      if 1, shuffle the buffer (i.e. list of processing files) and
          all the previous
      if 2, shuffle the returned batch and all the previous
  callbacks: None, or list of `odin.training.Callback`
      callback will be promoted during the execution of the task
  labels: None, or list of string
      labels for printing the confusion matrix in `odin.utils.Progbar`
  name: None or string
      unique name for Task identity.
  verbose : {0, 1, 2, 3, 4}
      specific verbose level controlling the log output
      0 - Turn off all log
      1 - progress off, only notification
      2 - progress off, notification and summary
      3 - progress on, nothing else
      4 - progress on, notification and summary
      5 - progress on, notification, summary and batch report
  """

  def __init__(self, func, data, epoch=1, p=1.0,
               batch_size=128, seed=None, shuffle_level=2,
               callbacks=None, labels=None, name=None,
               verbose=2):
    super(Task, self).__init__()
    self.set_func(func, data)
    # this Progbar will record the history as well
    self._labels = [str(l) for l in labels] \
        if labels is not None else None
    self._progbar = Progbar(target=self.nb_samples, name=name,
                            interval=0.,
                            print_report=True, print_summary=True)
    self._progbar.set_labels(self._labels)
    # ====== set callback and verbose ====== #
    self._callback = CallbackList(callbacks)
    self.set_verbose(verbose)
    # ====== assign other arguments ====== #
    self._nb_epoch = epoch
    self._p = np.clip(p, 0., 1.)
    self._seed = seed
    self.set_batch(batch_size, seed, shuffle_level)
    self._name = name
    # ====== current info ====== #
    self._curr_epoch = 0
    self._curr_iter = 0
    self._curr_samples = 0
    self._curr_epoch_iter = 0
    self._curr_epoch_samples = 0
    self._callback_msg = []
    # ====== iter tracking ====== #
    self._created_iter = None
    self._stop = False

  def __str__(self):
    return "<Task:'%s' p:%s bs:%s #ep:%s/%s #it:%s/%s #n:%s/%s %s>" % \
    (ctext(self.name, 'lightyellow'),
     ctext(self.probability, 'cyan'),
     ctext(self.batch_size, 'cyan'),
     ctext(self.curr_epoch, 'lightcyan'), ctext(self.nb_epoch, 'cyan'),
     ctext(self.curr_epoch_iter, 'lightcyan'), ctext(self.curr_iter, 'cyan'),
     ctext(self.curr_epoch_samples, 'lightcyan'), ctext(self.curr_samples, 'cyan'),
     ','.join([ctext(i.__class__.__name__, 'cyan')
               for i in self._callback._callbacks]))

  def __getstate__(self):
    return (self._progbar, self._nb_epoch, self._p, self._name,
            self._batch_size, self._rng, self._seed,
            self._shuffle_level, self._verbose)

  def __setstate__(self, states):
    (self._progbar, self._nb_epoch, self._p, self._name,
     self._batch_size, self._rng, self._seed,
     self._shuffle_level, self._verbose) = states
    # ====== current info ====== #
    self._curr_epoch = 0
    self._curr_iter = 0
    self._curr_samples = 0
    self._curr_epoch_iter = 0
    self._curr_epoch_samples = 0
    self._callback_msg = []
    # ====== iter tracking ====== #
    self._created_iter = None
    self._stop = False
    # ====== reset value of func and data ====== #
    self._func = None
    self._data = None

  def set_callbacks(self, callbacks):
    self._callback.set_callbacks(callbacks)
    if self._verbose == 0:
      self._callback.set_notification(False)
    else:
      self._callback.set_notification(True)
    return self

  def set_verbose(self, verbose):
    verbose = int(verbose)
    self._verbose = verbose
    if verbose == 0: # turn off everything
      self._callback.set_notification(False)
      self._progbar.print_progress = False
      self._progbar.print_summary = False
      self._progbar.print_report = False
    elif verbose == 1: # progress off, only notification
      self._callback.set_notification(True)
      self._progbar.print_progress = False
      self._progbar.print_summary = False
      self._progbar.print_report = False
    elif verbose == 2: # progress off, notification + summary
      self._callback.set_notification(True)
      self._progbar.print_progress = False
      self._progbar.print_summary = True
      self._progbar.print_report = False
    elif verbose == 3: # progress on, nothing else
      self._callback.set_notification(False)
      self._progbar.print_progress = True
      self._progbar.print_summary = False
      self._progbar.print_report = False
    elif verbose == 4: # progress on, notification + summary
      self._callback.set_notification(True)
      self._progbar.print_progress = True
      self._progbar.print_summary = True
      self._progbar.print_report = False
    elif verbose == 5: # progress on, notification, report, summary
      self._callback.set_notification(True)
      self._progbar.print_progress = True
      self._progbar.print_summary = True
      self._progbar.print_report = True
    else:
      raise ValueError(
          "Only support verbose value: 0, 1, 2, 3, 4, 5; but given: %s" % str(verbose))

  def set_func(self, func, data):
    # ====== check function ====== #
    self._func = func
    if isinstance(func, K.Function):
      self._output_info = [(o.name, o.shape.as_list())
                           for o in self._func.outputs]
    elif hasattr(func, '__call__'):
      self._output_info = [] # No info (normal function)
    else:
      raise ValueError("No support for function type: %s" %
          func.__class__.__name__)
    # ====== check data ====== #
    if not isinstance(data, (tuple, list)):
      data = [data]
    self._data = [fuel.as_data(i, copy=not isinstance(i, fuel.Feeder))
                  for i in data]
    self._nb_samples = min([d.iter_len for d in self._data])
    return self

  def set_batch(self, batch_size=None, seed=-1, shuffle_level=None):
    if batch_size is not None:
      self._batch_size = batch_size
    if seed is None or seed >= 0:
      if seed is not None:
        self._rng = np.random.RandomState(seed)
      else:
        self._rng = struct()
        self._rng.randint = lambda x: None
        self._rng.rand = get_rng().rand
    if shuffle_level is not None:
      self._shuffle_level = min(max(int(shuffle_level), 0), 2)
    return self

  # ==================== Properties ==================== #
  @property
  def history(self):
    """ Return : dictionary type
      {epoch_id : {tensor_name0: [batch_return1, batch_return2, ...],
                   tensor_name1: [batch_return1, batch_return2, ...],
                   ...},
       1 : {tensor_name0: [batch_return1, batch_return2, ...],
                  tensor_name1: [batch_return1, batch_return2, ...],
                  ...},
       ... }

    Example
    -------
    >>> for task_name, task_hist in task.history.items():
    >>>   print(task_name)
    >>>   for epoch_id, values in task_hist.items():
    >>>     print('  Epoch:', epoch_id)
    >>>     for tensor_name, v in values.items():
    >>>       print('  ', tensor_name, len(v))
    """
    return self._progbar.history

  @property
  def progbar(self):
    return self._progbar

  @property
  def name(self):
    return str(self._name)

  @property
  def labels(self):
    return self._labels

  @property
  def nb_epoch(self):
    return self._nb_epoch

  @property
  def nb_samples(self):
    ''' Estimated number of iteration for each epoch '''
    return self._nb_samples

  @property
  def probability(self):
    """Chance that the func will be execute during iteration"""
    return self._p

  @property
  def iter_per_epoch(self):
    ''' Estimated number of iteration for each epoch '''
    return int(np.ceil(self._nb_samples / self._batch_size))

  @property
  def batch_size(self):
    return self._batch_size

  @property
  def curr_epoch(self):
    """Total number of epoch finished since the beginning of the Task"""
    return self._curr_epoch

  @property
  def curr_iter(self):
    """Total number of iteration finished since the beginning of the Task"""
    return self._curr_iter

  @property
  def curr_samples(self):
    """Total number of samples finished since the beginning of the Task"""
    return self._curr_samples

  @property
  def curr_epoch_iter(self):
    """Number of iteration within current epoch"""
    return self._curr_epoch_iter

  @property
  def curr_epoch_samples(self):
    """Number of samples within current epoch"""
    return self._curr_epoch_samples

  @property
  def callback_msg(self):
    return self._callback_msg

  # ==================== control function ==================== #
  def stop(self):
    """ Stop all iterations running for this Task"""
    if self._created_iter is not None:
      self._stop = True
      # just run to end of the iterators
      for i in self._created_iter:
        pass
      self._stop = False
      self._created_iter = None

  def copy(self):
    return Task(self._func, self._data,
                epoch=self.nb_epoch, p=self.probability,
                batch_size=self.batch_size, seed=self._seed,
                shuffle_level=self._shuffle_level,
                name=self._name, verbose=self._verbose)

  def __iter(self):
    '''
    Return
    ------
    One of the following:
    * 'task_start':
    * 'epoch_start' : beginning of epoch
    * 'epoch_end' : epoch ended
    * 'task_end' : task ended
    * (results, nb_iter, nb_samples,
       nb_total_samples, nb_epoch) : results of execute function on data

    Note
    ----
    'end_task' also end of final epoch
    '''
    yield None # just for initalize the iterator
    self._callback_msg = self._callback.task_start(self)
    yield 'task_start'
    if self._stop:
      yield 'task_end'
    else:
      # ====== start of training ====== #
      while self._curr_epoch < self._nb_epoch:
        self._callback_msg = self._callback.epoch_start(self, self._data)
        yield 'epoch_start'
        seed = self._rng.randint(10e8)
        # if only 1 Data, don't need zip or we will mess up
        if len(self._data) == 1:
          data_it = iter(self._data[0].set_batch(batch_size=self._batch_size,
                                                 seed=seed,
                                                 shuffle_level=self._shuffle_level))
          data = data_it
        else:
          data_it = [iter(d.set_batch(batch_size=self._batch_size,
                                      seed=seed,
                                      shuffle_level=self._shuffle_level))
                     for d in self._data]
          data = zip(*data_it)
        # ======  start the iteration ====== #
        self._curr_epoch_samples = 0
        self._curr_epoch_iter = 0
        with self._progbar.safe_progress():
          for i, x in enumerate(data):
            # alread terminated, try to exhausted the iterator
            # if forced_to_terminate: continue
            # preprocessed the data
            if not isinstance(x, (tuple, list)):
              x = [x]
            # update some info
            shape0 = x[0].shape[0]
            self._curr_samples += shape0
            self._curr_iter += 1
            self._curr_epoch_samples += shape0
            self._curr_epoch_iter += 1
            self._callback_msg = self._callback.batch_start(self, x)
            # apply the function
            if self.probability >= 1. or self._rng.rand() < self.probability:
              results = self._func(*x)
              # add msg from batch_end event
              self._callback_msg += self._callback.batch_end(self, results)
              # return results
              yield results
              # update the progress bar
              for (name, shape), res in zip(self._output_info,
                                            as_tuple(results)):
                if len(shape) == 0: # return single value
                  self._progbar[name] = res
                else: # return tensor
                  self._progbar[name] = res
              self._progbar.add(shape0)
            # check TERMINATE signal
            if self._stop:
              # send signal to the data iterators also
              for i in data_it:
                if hasattr(i, 'stop'):
                  i.stop()
                else: # just iterate all over
                  for _ in i: pass
              # break the epoch loop
              break
        ### Epoch end signaling
        self._curr_epoch += 1
        self._callback_msg = self._callback.epoch_end(
            self, self._progbar.history[self._curr_epoch - 1])
        yield 'epoch_end'
        # ====== check if we got the right number for epoch iter ====== #
        if self._curr_epoch_samples != self._nb_samples:
          # just for sure should not smaller than the real number
          self._nb_samples = self._curr_epoch_samples
        # ======  end_epoch or task ====== #
        if self._stop or self._curr_epoch >= self._nb_epoch:
          self._callback_msg = self._callback.task_end(
              self, self._progbar.history)
          yield 'task_end'
          # showing notification
          if self._verbose >= 1 and self._verbose != 3:
            self._progbar.add_notification('Task "%s" ended!' % str(self.name))
          break
    # ====== end of iteration ====== #
    self._created_iter = None

  def __iter__(self):
    if self._created_iter is None:
      # reset all information
      self._curr_epoch = 0
      self._curr_iter = 0
      self._curr_samples = 0
      self._curr_epoch_iter = 0
      self._curr_epoch_samples = 0
      self._callback_msg = []
      # create new iter
      self._created_iter = self.__iter()
      # initialize the iteration
      next(self._created_iter)
    return self._created_iter

  def __del__(self):
    self.stop()


# ===========================================================================
# MainLoop
# ===========================================================================
class Timer(object):
  """Timer to determine when a `Task` should be start within a `MainLoop`

  Note
  ----
  The `Task` will be commenced if any of the conditions is `True`
  """

  def __init__(self, epoch=None, iteration=None, samples=None,
               percentage=None):
    super(Timer, self).__init__()
    self._epoch = epoch
    self._iteration = iteration
    self._samples = samples
    self._percentage = percentage
    # ====== for the counter ====== #
    self._counter_max = 0 # store maximum amount of sample
    self._counter = 0 # store current amount of samples
    self._last_samples_checkpoint = -1

  def check(self, task):
    # task haven't started
    if task.curr_samples == 0:
      return False
    if self._epoch is not None and task.curr_epoch >= self._epoch:
      return True
    if self._iteration is not None and \
    task.curr_iter >= self._iteration:
      return True
    if self._samples is not None and \
    task.curr_samples >= self._samples:
      return True
    if self._percentage is not None and\
    task.curr_samples / (task.nb_epoch * task.nb_samples) >= self._percentage:
      return True
    return False

  def set_counter(self, task):
    nb_samples = task.nb_samples
    batch_size = task.batch_size
    if self._epoch is not None:
      self._counter_max = nb_samples * self._epoch
    elif self._iteration is not None:
      self._counter_max = batch_size * self._iteration
    elif self._samples is not None:
      self._counter_max = self._samples
    elif self._percentage is not None:
      self._counter_max = int(self._percentage * nb_samples)
    self._counter = self._counter_max
    return self

  def update_counter(self, task):
    if self._last_samples_checkpoint < 0:
      self._last_samples_checkpoint = task.curr_samples
    else:
      self._counter -= task.curr_samples - self._last_samples_checkpoint
      self._last_samples_checkpoint = task.curr_samples
      if self._counter <= 0:
        self._counter = self._counter_max
        return True
    return False

class MainLoop(object):
  """
  Parameters
  ----------
  batch_size: int
      size of each batch return when iterate this Data
  seed: None, int
      if None, no shuffling is performed while iterating,
      if < 0, do not change the current seed
      if >= 0, enable randomization with given seed
  shuffle_level: int
      0: only shuffle the order of each batch
      1: shuffle the order of batches and inside each batch as well.
      2: includes level 0 and 1, and custom shuffling (strongest form)
  rollback: bool
      if True, rollback to the best checkpoint whenever the validation
      performance is degraded.
  verbose : {0, 1, 2, 3, 4}
      specific verbose level controlling the log output
      0 - Turn off all log
      1 - progress off, only notification
      2 - progress off, notification and summary
      3 - progress on, nothing else
      4 - progress on, notification and summary
      5 - progress on, notification, summary and batch report
  """

  def __init__(self, batch_size=256, seed=-1, shuffle_level=2,
               allow_rollback=True, labels=None,
               log_path=None, verbose=3):
    super(MainLoop, self).__init__()
    self._labels = labels
    self._main_task = None
    self._task = []
    self._subtask = []
    self._evaltask = []
    self._task_when = {} # mapping from `Task` to `Timer`
    self._task_freq = {} # mapping from `Task` to `Timer`
    self._allow_rollback = bool(allow_rollback)
    self._verbose = int(verbose)
    # create default RNG (no randomization)
    self._rng = struct()
    self._rng.randint = lambda *args, **kwargs: None
    # set batch
    self.set_batch(batch_size=batch_size, seed=seed,
                   shuffle_level=shuffle_level)
    self._callback = CallbackList()
    # ====== for the checkpoint ====== #
    self._save_path = None
    self._save_obj = None
    self._save_variables = []
    self._best_object = None
    self._save_history = True
    # ====== maximum stored checkpoint ====== #
    self._checkpoint_increasing = True
    self._checkpoint_max = -1
    self._current_checkpoint_count = 0

  # ==================== pickling ==================== #
  def __setstate__(self, value):
    self.set_batch(batch_size=value[0], shuffle_level=value[2])
    self._rng = value[1]

    self._callback = value[3]
    self._allow_rollback = value[4]

    self._task = []
    self._subtask = []
    self._main_task = None

  def __getstate__(self):
    raise NotImplementedError
    return (self._batch_size, self._rng, self._shuffle_level,
            self._callback, self._allow_rollback)

  # ==================== Signal handling ==================== #
  def _show_noti(self, msg):
    if self._verbose > 1 and self._verbose != 3:
      add_notification(msg)

  def set_checkpoint(self, path=None, obj=None, variables=[],
                     increasing=True, max_checkpoint=-1,
                     save_history=None):
    """ If `path` and `obj` given, the `obj` will be pickled
    at `path` for every checkpoint, otherwise, store the
    best values of `variables` in RAM

    Parameters
    ----------
    path: str
        path to save the obj when the callback return save signal

    obj: object
        any pickle-able object you want to save

    variables : {list of tensorflow.Variable}
        external variables will be saved together with the
        model

    increasing : bool (default: True)
        pass

    max_checkpoint : int (default: 3)
        pass
    """
    self._save_path = path
    self._save_obj = obj
    if variables is not None:
      variables = as_tuple(variables)
      if len(variables) > 0:
        from odin import backend as K
        self._save_variables = [v for v in variables if K.is_variable(v)]
    # other
    self._checkpoint_increasing = bool(increasing)
    self._checkpoint_max = int(max_checkpoint)
    # ====== get the latest checkpoint count ====== #
    if path is not None:
      saved_files = []
      base_dir = os.path.dirname(self._save_path)
      base_name = os.path.basename(self._save_path)
      pattern = re.compile(r"^%s\.\d+$" % re.escape(base_name))
      for name in os.listdir(base_dir):
        if not pattern.match(name):
          continue
        path = os.path.join(base_dir, name)
        if self._save_path + '.' in path:
          saved_files.append(path)
      saved_files = sorted(saved_files,
                           key=lambda x: int(x.split('.')[-1]))
      if len(saved_files) == 0:
        self._current_checkpoint_count = 0
      else:
        self._current_checkpoint_count = int(saved_files[-1].split('.')[-1]) + 1
      self._show_noti("[%s] Found lastest checkpoint: '.%d'" %
                      (ctext('MainLoop', 'red'), self._current_checkpoint_count))
    # ====== history ====== #
    if save_history is not None:
      self._save_history = bool(save_history)
    # save first checkpoint
    if self._current_checkpoint_count == 0:
      self._save(is_best=False)

  # ==================== properties ==================== #
  @property
  def labels(self):
    return self._labels

  @property
  def batch_size(self):
    return self._batch_size

  @property
  def history(self):
    """ Return : dictionary type
      {
      task1_name :
        {epoch_id : {tensor_name0: [batch_return1, batch_return2, ...],
                     tensor_name1: [batch_return1, batch_return2, ...],
                     ...},
         1 : {tensor_name0: [batch_return1, batch_return2, ...],
                    tensor_name1: [batch_return1, batch_return2, ...],
                    ...},
         ... },
      task2_name : ...
      }

    Example
    -------
    >>> for epoch_id, results in task.history.items():
    >>>   for tensor_name, values in results.items():
    >>>     print(tensor_name, len(values))
    """
    if self._main_task is None:
      return {}
    return OrderedDict([
        (t.name, t.history)
        for t in [self._main_task] + self._subtask + self._evaltask])

  def set_batch(self, batch_size=None, seed=-1, shuffle_level=None):
    """
    Parameters
    ----------
    batch_size: int
        size of each batch return when iterate this Data
    seed: None, int
        if None, no shuffling is performed while iterating,
        if < 0, do not change the current seed
        if >= 0, enable randomization with given seed
    start: int, float
        if int, start indicates the index of starting data points to
        iterate. If float, start is the percentage of data to start.
    end: int, float
        ending point of the interation
    shuffle_level: int
        0: only shuffle the order of each batch
        1: shuffle the order of batches and inside each batch as well.
        2: includes level 0 and 1, and custom shuffling (strongest form)
    """
    if batch_size is not None:
      self._batch_size = batch_size
    if seed is None or seed >= 0:
      if seed is not None:
        self._rng = np.random.RandomState(seed)
      else:
        self._rng = struct()
        self._rng.randint = lambda *args, **kwargs: None
    if shuffle_level is not None:
      shuffle_level = min(max(int(shuffle_level), 0), 2)
      self._shuffle_level = shuffle_level
    # ====== set_batch for Tasks ====== #
    for i in self._task:
      i.set_batch(batch_size=batch_size, seed=seed,
                  shuffle_level=shuffle_level)
    for i in self._subtask:
      i.set_batch(batch_size=batch_size, seed=seed,
                  shuffle_level=shuffle_level)
    return self

  @property
  def callback(self):
    return self._callback

  def set_callbacks(self, callbacks):
    self._callback.set_callbacks(callbacks)
    return self

  def __getitem__(self, key):
    """ Return callback from callback list"""
    return self._callback[key]

  # ==================== main ==================== #
  def set_train_task(self, func, data, epoch=1, p=1., when=None,
                     name="Train"):
    ''' The progress of the first task added as train_task will
    be used to determine Timer for all other task

    Parameters
    ----------
    '''
    if len(self._task) == 0 and when is not None:
      raise ValueError("First train task will always be executed, you cannot"
          "specify `Timer` for this task.")
    if when is not None and not isinstance(when, Timer):
      raise ValueError("`when` must be instance of odin.training.Timer")
    t = Task(func, data, epoch=epoch, p=p,
             batch_size=self._batch_size,
             seed=self._rng.randint(10e8),
             shuffle_level=self._shuffle_level,
             labels=self.labels, name=name,
             verbose=self._verbose)
    self._task.append(t)
    self._task_when[t] = when
    self._task_freq[t] = Timer(samples=0)
    # assign main task
    if self._main_task is None:
      self._main_task = t
    return self

  def set_valid_task(self, func, data,
                     freq=Timer(epoch=1), when=Timer(samples=0),
                     name="Valid"):
    ''' A subtask is a repeative Task

    Parameters
    ----------
    when: {Timer}
        percentage of epoch of main task before this task is executed
        negative value => execute after final epoch of main task
    freq: {Timer}
        percentage of epoch of main task before this task is executed
    '''
    if not isinstance(when, Timer):
      raise ValueError("`when` must be instance of odin.training.Timer")
    t = Task(func, data, epoch=float('inf'), p=1.,
             batch_size=self._batch_size,
             seed=None, shuffle_level=0,
             labels=self.labels, name=name,
             verbose=self._verbose)
    self._subtask.append(t)
    self._task_when[t] = when
    self._task_freq[t] = freq
    return self

  def set_eval_task(self, func, data, name="Eval", labels=None):
    t = Task(func, data, epoch=1, p=1., batch_size=self._batch_size,
             seed=None, shuffle_level=0,
             labels=labels, name=name,
             verbose=self._verbose)
    self._evaltask.append(t)
    self._task_when[t] = Timer(percentage=1.)
    self._task_freq[t] = Timer(samples=0)
    return self

  # ==================== logic ==================== #
  def _save(self, is_best):
    is_best = bool(is_best)
    # trigger event for callbacks
    self._callback.event(TrainSignal.SAVE_BEST
                         if is_best else
                         TrainSignal.SAVE)
    # ====== save the model to hard drive ====== #
    if self._save_path is not None:
      # serialize the best model to disk
      if is_best:
        final_save_path = self._save_path
        N.serialize(nnops=self._save_obj, path=self._save_path,
                    save_variables=True, variables=self._save_variables,
                    binary_output=False, override=True)
      # not the best model saved, just periodically saving
      else:
        final_save_path = self._save_path + '.%d' % self._current_checkpoint_count
        N.serialize(nnops=self._save_obj,
                    path=final_save_path,
                    save_variables=True, variables=self._save_variables,
                    binary_output=False, override=True)
        self._current_checkpoint_count += 1
        if self._checkpoint_max > 1 and self._current_checkpoint_count > self._checkpoint_max:
          shutil.rmtree(
              self._save_path + '.%d' %
              (self._current_checkpoint_count - self._checkpoint_max - 1))
      # print the log
      self._show_noti("[%s] Creating %scheckpoint at: %s" %
                      (ctext('MainLoop', 'red'),
                       ctext('[best]', 'yellow') if is_best else '',
                       final_save_path))
      # save history
      if self._save_history:
        with open(final_save_path + '.hist', 'wb') as f:
          pickle.dump(self.history, f)
        self._show_noti("[%s] Save history at: %s" %
                        (ctext('MainLoop', 'red'),
                         final_save_path + '.hist'))
    # ====== store the object directly in RAM (only for the best) ====== #
    elif bool(is_best) and \
    (self._save_obj is not None or len(self._save_variables) > 0):
      del self._best_object
      self._best_object = N.serialize(
          self._save_obj, path=None, save_variables=True,
          variables=self._save_variables, binary_output=True)
      mem_size = sum(len(v) for k, v in self._best_object.items()) / 1024 / 1024
      self._show_noti(
          "[%s] Creating dynamic checkpoint in RAM using %.2f (megabytes)" %
          (ctext('MainLoop', 'red'), mem_size))

  def _rollback(self, is_final=False):
    # TODO: update rollback mechanism
    if not self._allow_rollback and not is_final:
      return
    # trigger event for callbacks
    self._callback.event(TrainSignal.ROLLBACK)
    # default rollback procedure
    if self._save_path is not None and os.path.exists(self._save_path):
      self._show_noti("[%s] Rollback from: %s" %
                      (ctext('MainLoop', 'red'), self._save_path))
      # restore previous checkpoint immediately
      N.deserialize(self._save_path, force_restore_vars=True)
    # otherwise, load stored variables from RAM
    elif self._best_object is not None:
      self._show_noti("[%s] Rollback to the best stored object from RAM" %
                      (ctext('MainLoop', 'red')))
      N.deserialize(path_or_data=self._best_object,
                    force_restore_vars=True)

  def _run(self):
    if self._main_task is None and len(self._evaltask) == 0:
      raise ValueError('You must call `set_task` and set the main task '
          'first, or you can specify evaluation task using `set_eval_task`.')
    # ====== set the callback for all Task, make sure all Callbacks are up-to-date ====== #
    for t in self._task + self._subtask + self._evaltask:
      t.set_callbacks(self._callback)
      t.set_verbose(self._verbose)
    # ====== prepare subtask ====== #
    if self._main_task is not None:
      finished_task = {i: False for i in self._task + self._subtask}
      task_iter = {i: iter(i) for i in self._task + self._subtask}
      for freq in self._task_freq.values():
        freq.set_counter(self._main_task)
    # ====== main logics ====== #
    while self._main_task is not None and \
    not finished_task[self._main_task]:
      for t in self._task:
        # ====== execute training task first ====== #
        if not finished_task[t]:
          res = next(task_iter[t])
          # task message
          if isinstance(res, str):
            if res == 'task_end':
              finished_task[t] = True
          else: # task result
            pass
          # process callback msg for tasks
          msg = t.callback_msg
          if TrainSignal.SAVE in msg:
            self._save(is_best=False)
          if TrainSignal.SAVE_BEST in msg:
            self._save(is_best=True)
          if TrainSignal.ROLLBACK in msg:
            self._rollback()
          if TrainSignal.STOP in msg:
            self._callback.event(TrainSignal.STOP)
            finished_task[self._main_task] = True
            break
        # ====== execute valid and eval task ====== #
        for st in self._subtask:
          if finished_task[self._main_task]:
            break
          if finished_task[st]:
            continue
          if self._task_when[st].check(self._main_task) and \
          self._task_freq[st].update_counter(self._main_task):
            # running 1 epoch of subtask
            for x in task_iter[st]:
              # process callback msg for subtasks
              msg = st.callback_msg
              if TrainSignal.SAVE in msg:
                self._save(is_best=False)
              if TrainSignal.SAVE_BEST in msg:
                self._save(is_best=True)
              if TrainSignal.ROLLBACK in msg:
                self._rollback()
              if TrainSignal.STOP in msg:
                self._callback.event(TrainSignal.STOP)
                finished_task[self._main_task] = True
                break

              # signal
              if isinstance(x, str):
                if x == 'task_end': finished_task[st] = True
                if x == 'epoch_end': break
              else: # results
                pass
    # ====== end main task ====== #
    for t in self._task + self._subtask:
      t.stop()
    # ====== Run eval task before finishing ====== #
    # rollback to the best check point for final evaluation
    self._rollback(is_final=True)
    for et in self._evaltask:
      for x in iter(et):
        pass
    # everything finished
    self._evaltask = []

  def run(self):
    try:
      self._run()
    finally:
      try:
        import curses
        curses.echo()
        curses.nocbreak()
        curses.endwin()
      except Exception:
        pass



from __future__ import print_function, division, absolute_import

import types
import inspect
from collections import Mapping
from contextlib import contextmanager

import numpy as np
import tensorflow as tf

from odin import backend as K
from odin.nnet.base import Container
from odin.utils import (as_tuple, is_number, flatten_list, ctext,
                        axis_normalize)

# ===========================================================================
# Implementation
# ===========================================================================
class Residual(Container):
  pass

class StochasticDepth(Container):
  pass

class TimeDistributed(Container):
  """ Time distributed container applying NNOp or list of NNOp
  along given `time_axis` of input tensor

  Parameters
  ----------
  """

  def __init__(self, ops, time_axis=1, backward=False, reverse=False,
               debug=False, **kwargs):
    super(TimeDistributed, self).__init__(**kwargs)
    self.set_nnops(ops)
    self.time_axis = int(time_axis)
    self.backward = bool(backward)
    self.reverse = bool(reverse)
    self.debug = int(debug)

  def _apply(self, X, mask=None):
    def _step_fn(outs, ins):
      return [f(ins) for f in self._apply_ops]
    # ====== need to apply the ops to know initializer information ====== #
    ndim = X.shape.ndims
    axis = axis_normalize(self.time_axis, ndim=ndim)
    with tf.device("/cpu:0"):
      sample = tf.zeros_like(X)
      sample = sample[[slice(None, None) if i != axis else 0
                       for i in range(ndim)]]
    initializer = [tf.zeros_like(f(sample)) for f in self._apply_ops]
    # ====== scan ====== #
    outputs = K.scan_tensors(_step_fn,
                             sequences=X, mask=mask, initializer=initializer,
                             axis=axis,
                             backward=self.backward, reverse=self.reverse,
                             reshape_outputs=True)
    return outputs[0] if len(self._apply_ops) == 1 else outputs

  def _transpose(self):
    return TimeDistributed(ops=[o.T for o in self._apply_ops],
                           time_axis=self.time_axis,
                           backward=self.backward,
                           reverse=self.reverse)

class Parallel(Container):
  """ Parallel

  Parameters
  ----------
  mode : {None, 'concat', 'max', 'min', 'sum', 'mean'}
    None - no post-processing is performed, return the outputs from all NNOp
    'concat' - concatenate the outputs along given `axis`
    'min', 'max' - take the min or max values of the outputs along given `axis`
    'sum' - take sum of the outputs
    'mean' - take mean of the outputs
  """

  def __init__(self, ops, mode='concat', axis=-1,
               debug=False, **kwargs):
    super(Parallel, self).__init__(**kwargs)
    self.set_nnops(ops)
    mode = str(mode).lower()
    assert mode in ('none', 'concat', 'min', 'max', 'sum', 'mean'),\
    "Support `mode` includes: 'none', 'concat', 'min', 'max', 'sum', 'mean';" +\
    " but given: %s" % mode
    self.mode = mode
    self.axis = axis
    self.debug = int(debug)

  def _apply(self, *args, **kwargs):
    with self._debug_mode(*args, **kwargs):
      outputs = []
      for op in self._apply_ops:
        outputs.append(op(*args, **kwargs))
        self._print_op(op)
      # ====== post-processing ====== #
      if self.mode == 'concat':
        ret = tf.concat(outputs, self.axis)
      elif self.mode == 'max':
        ret = outputs[0]
        for o in outputs[1:]:
          ret = tf.maximum(ret, o)
      elif self.mode == 'min':
        ret = outputs[0]
        for o in outputs[1:]:
          ret = tf.minimum(ret, o)
      elif self.mode == 'sum' or self.mode == 'mean':
        outputs[0]
        for o in outputs[1:]:
          ret = ret + o
        if self.mode == 'mean':
          ret = ret / len(outputs)
      else:
        ret = outputs
    return ret

  def _transpose(self):
    pass

class Sequence(Container):

  """ Sequence of Operators

  Parameters
  ----------
  strict_transpose: bool
      if True, only operators with transposed implemented are added
      to tranpose operator
  debug: bool
      if `1`, print NNOp name and its input and output shape
      if `2`, print all information of each NNOp
  return_all_layers: bool
      if True, return the output from all layers instead of only the last
      layer.

  Note
  ----
  You can specify kwargs for a specific NNOp using `params` keywords,
  for example: `params={1: {'noise': 1}}` will applying noise=1 to the
  1st NNOp, or `params={(1, 2, 3): {'noise': 1}}` to specify keywords
  for the 1st, 2nd, 3rd Ops.

  Example
  -------
  """

  def __init__(self, ops, return_all_layers=False,
               strict_transpose=False, debug=False, **kwargs):
    super(Sequence, self).__init__(**kwargs)
    # ====== validate ops list ====== #
    self.set_nnops(ops)
    self.return_all_layers = bool(return_all_layers)
    self.strict_transpose = bool(strict_transpose)
    self.debug = int(debug)

  def _apply(self, *args, **kwargs):
    with self._debug_mode(*args, **kwargs):
      all_outputs = []
      for i, op in enumerate(self._apply_ops):
        if i == 0:
          x = op(*args, **kwargs)
        else:
          x = op(x, **kwargs)
        all_outputs.append(x)
        # print after finnish the op
        self._print_op(op)
    return all_outputs if self.return_all_layers else x

  def _transpose(self):
    transpose_ops = []
    for i in self._apply_ops:
      if hasattr(i, 'T'):
        transpose_ops.append(i.T)
      elif not self.strict_transpose:
        transpose_ops.append(i)
    # reversed the order of ops for transpose
    transpose_ops = list(reversed(transpose_ops))
    seq = Sequence(transpose_ops, debug=self.debug,
                   name=self.name + '_transpose')
    return seq



# ===========================================================================
# Available properties
#  * device
#  * device_info
#  * floatX
#  * epsilon
#  * multigpu
#  * optimizer
#  * cnmem
#  * backend
#  * seed
# ===========================================================================
from __future__ import division, absolute_import

import os
import re
import sys
import pip
import shutil
import tempfile
import subprocess
import warnings
from six import string_types
from multiprocessing import cpu_count
try:
  from functools import lru_cache
except ImportError:
  def lru_cache(maxsize=128):
    def tmp_func(func):
      return func
    return tmp_func

import numpy as np

# ===========================================================================
# Helper
# ===========================================================================
def _ctext(s, color='red'):
  # just a copy of ctext implementation to make
  # the config log look better.
  try:
    from colorama import Fore
    color = color.upper()
    color = getattr(Fore, color, '')
    return color + str(s) + Fore.RESET
  except ImportError:
    pass
  return str(s)

def _warning(text):
  print(_ctext('[WARNING]', 'red'), text)

def _check_package_available(name):
  for i in pip.get_installed_distributions():
    if name.lower() == i.key.lower():
      return True
  return False

def get_gpu_info():
  """ Example of return
  [
    {'name': 'Titan TITAN V', 'id': 0,
     'fan': 53,
     'mem_total': 12028, 'mem_used': 4295,
     'graphics_clock': 1200, 'mem_clock': 850
    },
    {'name': 'GeForce TITAN Xp', 'id': 1,
    'fan': 23,
    'mem_total': 12196, 'mem_used': 0,
    'graphics_clock': 1404, 'mem_clock': 5705
    }
  ]
  """
  temp_dir = tempfile.mkdtemp()
  p = os.path.join(temp_dir, 'tmp.txt')
  queried = subprocess.call('nvidia-smi -q -x > ' + p,
          shell=True,
          stdout=subprocess.PIPE,
          stderr=subprocess.PIPE) == 0
  devices = []
  if queried: # found nvidia-smi
    from xml.etree import ElementTree as ET
    tree = ET.parse(p)
    root = tree.getroot()
    for child in root:
      if child.tag.lower() != 'gpu':
        continue
      gpu = {}
      brand = ''
      for i in child:
        tag = i.tag.lower()
        if tag == 'product_name':
          gpu['name'] = i.text
        elif tag == 'product_brand':
          brand = i.text
        elif tag == 'minor_number':
          gpu['id'] = int(i.text)
        elif tag == 'fan_speed':
          gpu['fan'] = int(i.text.split(' ')[0])
        elif tag == 'fb_memory_usage':
          gpu['mem_total'] = int(i.findall('total')[0].text.split(' ')[0])
          gpu['mem_used'] = int(i.findall('used')[0].text.split(' ')[0])
        elif tag == 'clocks':
          for j in i:
            gpu[j.tag] = int(j.text.split(' ')[0])
      gpu['name'] = brand + ' ' + gpu['name']
      devices.append(gpu)
    # sort by minor number
    devices = sorted(devices, key=lambda x: x['id'])
  else: # NO GPU devices
    pass
  # remove temp-dir
  shutil.rmtree(temp_dir)
  return devices

class AttributeDict(dict):
  __getattr__ = dict.__getitem__
  __setattr__ = dict.__setitem__

  def __setstate__(self, states):
    for i, j in states:
      self[i] = j

  def __getstate__(self):
    return self.items()

# ===========================================================================
# Auto config
# ===========================================================================
CONFIG = None
EPS = None
_SESSION = {}
_RNG_GENERATOR = None

def set_session(session):
  global _SESSION
  _SESSION = session

def get_session_config():
  import tensorflow as tf
  session_args = {
      'intra_op_parallelism_threads': CONFIG['nthread'],
      'inter_op_parallelism_threads': CONFIG['ncpu'],
      'allow_soft_placement': True,
      'log_device_placement': CONFIG['debug'],
  }
  if CONFIG['ngpu'] > 0:
    if CONFIG['cnmem'] > 0:
      session_args['gpu_options'] = tf.GPUOptions(
          per_process_gpu_memory_fraction=CONFIG['cnmem'],
          allow_growth=False)
    else:
      session_args['gpu_options'] = tf.GPUOptions(allow_growth=True)
  return session_args

def get_session(graph=None):
  """ Calling this method will make sure you create
  only 1 Session per graph.
  """
  # avoid duplicate Session for the same Graph when
  # graph is None
  if graph is None:
    import tensorflow as tf
    default_graph = tf.get_default_graph()
    if default_graph in _SESSION:
      _SESSION[None] = _SESSION[default_graph]
  # another case
  elif None in _SESSION and _SESSION[None].graph == graph:
    _SESSION[graph] = _SESSION[None]
  # ====== initialize tensorflow session ====== #
  if graph not in _SESSION:
    import tensorflow as tf
    session_args = get_session_config()
    sess = tf.Session(config=tf.ConfigProto(**session_args), graph=graph)
    _SESSION[graph] = sess
    with sess.graph.as_default():
      tf.Variable(initial_value=False, dtype=tf.bool, name='IsTraining__',
                  trainable=False)
  return _SESSION[graph]

def auto_config(config=None):
  ''' Auto-configure ODIN using os.environ['ODIN'].

  Parameters
  ----------
  config : object
    predefined config object return from auto_config
  check : boolean
    if True, raise Exception of CONFIG have not initialized

  Returns
  -------
  config : object
    simple object (kind of dictionary) contain all configuraitons

  '''
  global CONFIG
  global _RNG_GENERATOR
  global EPS
  if CONFIG is not None:
    warnings.warn('You should not auto_config twice, old configuration already '
        'existed, and cannot be re-configured.')
    return CONFIG
  # ====== specific pattern ====== #
  valid_cnmem_name = re.compile(r"(cnmem)[=]?[10]?\.\d*")
  valid_seed = re.compile(r"seed\D?(\d*)")

  floatX = 'float32'
  epsilon = 1e-8
  cnmem = 0.
  seed = 1208251813
  debug = False
  # number of devices
  ncpu = 0
  nthread = 0
  log_level = '3'

  # Handling GPU devices
  #  * Nothing => "0" => use first GPU found
  #  * "gpu" => "-1" => use all available GPU devices
  #  * "gpu=" => "" => remove all GPU devices
  #  * "gpu=0_1" => "0,1" => use 0-th and 1-st GPU
  #  * "gpu=0_1_-1" => "-1" => use all available GPU devices
  ngpu = "0"
  # ====== parsing the config ====== #
  if config is None: # load config from flags
    odin_flags = os.getenv("ODIN", "")
    for c in (';', ':', '.', '*'):
      odin_flags.replace(c, ',')
    for c in ('+', '_'):
      odin_flags.replace(c, '_')
    s = odin_flags.split(',')
    # ====== processing each tag ====== #
    for i in s:
      i = i.lower().strip()
      # ====== Data type ====== #
      if 'float' in i or 'int' in i:
        floatX = i
      # ====== Devices ====== #
      elif 'cpu' in i: # CPU devices
        if '=' not in i:
          _warning("Found `cpu` tag, but number of CPU is not "
             "specified, auto select all %d CPU-cores " %
             cpu_count())
        else:
          ncpu = min(int(i.split('=')[-1]), cpu_count())
      elif 'gpu' in i: # GPU devices
        # given a set of GPU devices, multiple devices specified by 0,1,2,3
        if '=' in i:
          ngpu = i.split('=')[-1]
          if len(ngpu) > 0:
            try:
              [int(_) for _ in ngpu.split('_')]
            except Exception as e:
              raise RuntimeError("Invalid input for GPU configuration")
            if any(int(_) < 0 for _ in ngpu.split('_')):
              ngpu = "-1"
        # use all available GPU devices
        else:
          ngpu = "-1"
      # ====== number thread ====== #
      elif 'thread' in i:
        if '=' not in i:
          _warning("Found `thread` tag, but number of thread is "
             "not specified, only 1 thread per process (CPU-core) is "
             "used by default.")
        else:
          nthread = int(i.split('=')[-1])
      # ====== cnmem ====== #
      elif 'cnmem' in i:
        match = valid_cnmem_name.match(i)
        if match is None:
          raise ValueError('Unsupport CNMEM format: %s. '
               'Valid format must be: cnmem=0.75 or cnmem=.75 '
               ' or cnmem.75' % str(i))
        i = i[match.start():match.end()].replace('cnmem', '').replace('=', '')
        cnmem = float(i)
      # ====== seed ====== #
      elif 'seed' in i:
        match = valid_seed.match(i)
        if match is None:
          raise ValueError('Invalid pattern for specifying seed value, '
               'you can try: [seed][non-digit][digits]')
        seed = int(match.group(1))
      # ====== debug ====== #
      elif 'debug' in i:
        debug = True
      # ====== log-leve ====== #
      elif 'log' in i:
        if '=' in i:
          log_level = i.split('=')[-1]
        else:
          log_level = '0'
  else: # load config from object
    floatX = config['floatX']
    ncpu = config['ncpu']
    ngpu = config['ngpu']
    nthread = config['nthread']
    cnmem = config['cnmem']
    seed = config['seed']
    debug = config['debug']
  # ====== set log level ====== #
  # This only worked if I put the os.environ before tensorflow was imported
  # 0 = all messages are logged (default behavior)
  # 1 = INFO messages are not printed
  # 2 = INFO and WARNING messages are not printed
  # 3 = INFO, WARNING, and ERROR messages are not printed
  os.environ['TF_CPP_MIN_LOG_LEVEL'] = log_level
  # ====== get device info ====== #
  # epsilon
  EPS = np.finfo(np.dtype(floatX)).eps
  # devices
  dev = {}
  dev['ncpu'] = ncpu
  dev['nthread'] = nthread
  # ==================== processing GPU devices ==================== #
  gpu_info = get_gpu_info()
  if len(ngpu) > 0:
    if ngpu == '-1':
      dev['ngpu'] = len(gpu_info)
    else:
      dev['ngpu'] = min(len(gpu_info), len(ngpu.split('_')))
  else:
    dev['ngpu'] = 0
  #
  if dev['ngpu'] == 0:
    os.environ['CUDA_VISIBLE_DEVICES'] = ""
    gpu_info = []
  elif ngpu == '-1':
    os.environ['CUDA_VISIBLE_DEVICES'] = ",".join(
        [str(i) for i in range(len(gpu_info))])
  else:
    all_gpu = sorted([_
                      for _ in ngpu.split('_')
                      if int(_) < len(gpu_info)])
    os.environ['CUDA_VISIBLE_DEVICES'] = ",".join(all_gpu)
    gpu_info = [gpu_info[int(i)] for i in all_gpu]
  dev['gpu'] = gpu_info
  dev['gpu_indices'] = [
      int(i)
      for i in str(os.environ['CUDA_VISIBLE_DEVICES']).split(',')
      if len(i) > 0 and i.isdigit()]

  # ====== Log the configuration ====== #
  def print_log(tag, value, nested=False):
    if nested:
      s = _ctext('  * ', 'MAGENTA')
    else:
      s = '[Auto-Config] '
    s += _ctext(tag, 'yellow') + ' : '
    s += _ctext(value, 'cyan') + '\n'
    sys.stderr.write(s)

  print_log('#CPU', 'auto' if dev['ncpu'] == 0 else dev['ncpu'])
  print_log('#CPU-native', str(cpu_count()))
  print_log('#Thread/core',
      'auto' if dev['nthread'] == 0 else dev['nthread'],
      nested=True)
  print_log('#GPU', dev['ngpu'])
  if dev['ngpu'] > 0:
    for i in dev['gpu']:
      print_log('[%d]' % i['id'], i['name'], nested=True)
  print_log('FloatX', floatX)
  print_log('Epsilon', epsilon)
  print_log('CNMEM', cnmem)
  print_log('SEED', seed)
  print_log('Log-devices', debug)
  print_log('TF-log-level', log_level)
  # ====== Return global objects ====== #
  CONFIG = AttributeDict()
  CONFIG.update({
      'ncpu': dev['ncpu'],
      'ngpu': dev['ngpu'],
      'nthread': dev['nthread'],
      'gpu_indices': dev['gpu_indices'],
      'device_info': dev,
      'floatX': floatX,
      'epsilon': epsilon,
      'cnmem': cnmem,
      'seed': seed,
      'debug': debug
  })
  _RNG_GENERATOR = np.random.RandomState(seed=seed)
  with warnings.catch_warnings():
    warnings.filterwarnings(action='ignore', category=ImportWarning)
    import tensorflow as tf
    the_seed = _RNG_GENERATOR.randint(0, 10e8)
    tf.set_random_seed(seed=the_seed)
    np.random.seed(seed=the_seed)
  # ====== set the random seed for everything ====== #
  return CONFIG

# ===========================================================================
# Getter
# ===========================================================================
def __validate_config():
  if CONFIG is None:
    raise Exception("auto_config has not been called.")

def get_rng():
  """
    Return
    ------
    the numpy RandomState as a "Randomness Generator"
  """
  global _RNG_GENERATOR
  if _RNG_GENERATOR is None:
    _RNG_GENERATOR = np.random.RandomState(seed=120825)
  return _RNG_GENERATOR

get_random_state = get_rng

def randint(low=0, high=10e8, size=None, dtype='int32'):
  """Randomly generate and integer seed for any stochastic function."""
  return _RNG_GENERATOR.randint(low=low, high=high, size=size,
    dtype=dtype)

def get_ncpu():
  """ Return number of inter_op_parallelism_threads """
  __validate_config()
  return CONFIG['ncpu']

def get_ncpu_native():
  return cpu_count()

def get_ngpu():
  """ Return number of GPU """
  __validate_config()
  return CONFIG['ngpu']

def get_gpu_indices():
  __validate_config()
  return CONFIG['gpu_indices']

def get_nthread():
  """ Return number of intra_op_parallelism_threads """
  __validate_config()
  return CONFIG['nthread']

def get_nb_processors():
  """ In case using CPU, return number of cores
  If GPU is used, return number of graphics card.
  """
  __validate_config()
  return CONFIG['device_info']['n']

def get_device_info():
  """ Device info contains:
  {"n": ngpu,
   "dev0": [device_name, device_compute_capability, device_total_memory],
   "dev1": [device_name, device_compute_capability, device_total_memory],
   ...
  }
  """
  __validate_config()
  return CONFIG['device_info']

def get_floatX():
  __validate_config()
  return CONFIG['floatX']

def get_optimizer():
  __validate_config()
  return CONFIG['optimizer']

def get_cnmem():
  __validate_config()
  return CONFIG['cnmem']

def get_backend():
  __validate_config()
  return CONFIG['backend']

def get_random_seed():
  __validate_config()
  return CONFIG['seed']



from __future__ import print_function, division, absolute_import

import numpy as np
from scipy.linalg import eigh, cholesky, inv, svd, solve
import tensorflow as tf

from sklearn.svm import SVC
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis

from odin.backend import length_norm, calc_white_mat
from odin.ml.base import BaseEstimator, TransformerMixin, Evaluable

# ===========================================================================
# Cosine Scoring
# ===========================================================================
def compute_class_avg(X, y, classes, sorting=True):
  """ compute average vector for each class

  Parameters
  ----------
  X: [nb_samples, feat_dim]
  y: [nb_samples]
  classes: [nb_classes]
    assumed numerical classes
  sorting: bool
    if True, sort the `classes` by numerical order (from small to large)

  Return
  ------
  [nb_classes, feat_dim]

  Note
  ----
  The given order of each class in `classes` will determine
  the row order of returned matrix
  """
  if sorting:
    classes = sorted(classes, reverse=False)
  return np.concatenate([np.mean(X[y == i], axis=0, keepdims=True)
                         for i in classes],
                        axis=0)

def compute_within_cov(X, y, classes=None, class_avg=None):
  """ Compute the within-classes covariance matrix

  Parameters
  ----------
  X : [nb_samples, feat_dim]
  y : [nb_samples]
  classes : [nb_classes]
    assumed numerical classes
  class_avg : [nb_classes, feat_dim]
    concatenated average vector of each class

  Return
  ------
  [feat_dim, feat_dim]

  Note
  ----
  The given order of each class in `classes` will determine
  the row order of returned matrix
  """
  if classes is None and class_avg is None:
    raise ValueError("`classes` and `class_avg` cannot be None together")
  if classes is not None:
    class_avg = compute_class_avg(X, y, classes, sorting=True)
  X_mu = X - class_avg[y]
  Sw = np.cov(X_mu.T)
  return Sw

def compute_wccn(X, y, classes=None, class_avg=None):
  """ Within class covariance normalization

  Parameters
  ----------
  X : [nb_samples, feat_dim]
  y : [nb_samples]
  classes : [nb_classes]
    assumed numerical classes
  class_avg : [nb_classes, feat_dim]
    concatenated average vector of each class

  Return
  ------
  w: [feat_dim, feat_dim]
    where X_norm = dot(X, w)
  """
  if classes is None and class_avg is None:
    raise ValueError("`classes` and `class_avg` cannot be None together")
  Sw = compute_within_cov(X, y, classes, class_avg)
  Sw = Sw + 1e-6 * np.eye(Sw.shape[0])
  return calc_white_mat(Sw)

class VectorNormalizer(BaseEstimator, TransformerMixin):
  """ Perform of sequence of normalization as following
    -> Centering: Substract sample mean
    -> Whitening: using within-class-covariance-normalization
    -> Applying LDA (optional)
    -> Length normalization

  Parameters
  ----------
  centering : bool (default: True)
    mean normalized the vectors
  wccn : bool (default: True)
    within class covariance normalization
  lda : bool (default: True)
    Linear Discriminant Analysis
  concat : bool (default: False)
    concatenate original vector to the normalized vector

  Return
  ------
  [nb_samples, feat_dim] if `lda=False`
  [nb_samples, nb_classes - 1] if `lda=True` and `concat=False`
  [nb_samples, feat_dim + nb_classes - 1] if `lda=True` and `concat=True`

  """

  def __init__(self, centering=True, wccn=False, unit_length=True,
               lda=False, concat=False):
    super(VectorNormalizer, self).__init__()
    self._centering = bool(centering)
    self._unit_length = bool(unit_length)
    self._wccn = bool(wccn)
    self._lda = LinearDiscriminantAnalysis() if bool(lda) else None
    self._feat_dim = None
    self._concat = bool(concat)

  # ==================== properties ==================== #
  @property
  def feat_dim(self):
    return self._feat_dim

  @property
  def is_initialized(self):
    return self._feat_dim is not None

  @property
  def is_fitted(self):
    return hasattr(self, '_W')

  @property
  def enroll_vecs(self):
    return self._enroll_vecs

  @property
  def mean(self):
    """ global mean vector """
    return self._mean

  @property
  def vmin(self):
    return self._vmin

  @property
  def vmax(self):
    return self._vmax

  @property
  def W(self):
    return self._W

  @property
  def lda(self):
    return self._lda

  # ==================== sklearn ==================== #
  def _initialize(self, X, y):
    if not self.is_initialized:
      self._feat_dim = X.shape[1]
    assert self._feat_dim == X.shape[1]
    if isinstance(y, (tuple, list)):
      y = np.asarray(y)
    if y.ndim == 2:
      y = np.argmax(y, axis=-1)
    return y, np.unique(y)

  def normalize(self, X, concat=None):
    """
    Parameters
    ----------
    X : array [nb_samples, feat_dim]
    concat : {None, True, False}
      if not None, override the default `concat` attribute of
      this `VectorNormalizer`
    """
    if not self.is_fitted:
      raise RuntimeError("VectorNormalizer has not been fitted.")
    if concat is None:
      concat = self._concat
    if concat:
      X_org = X[:] if not isinstance(X, np.ndarray) else X
    else:
      X_org = None
    # ====== normalizing ====== #
    if self._centering:
      X = X - self._mean
    if self._wccn:
      X = np.dot(X, self.W)
    # ====== LDA ====== #
    if self._lda is not None:
      X_lda = self._lda.transform(X) # [nb_classes, nb_classes - 1]
      # concat if necessary
      if concat:
        X = np.concatenate((X_lda, X_org), axis=-1)
      else:
        X = X_lda
    # ====== unit length normalization ====== #
    if self._unit_length:
      X = length_norm(X, axis=-1, ord=2)
    return X

  def fit(self, X, y):
    y, classes = self._initialize(X, y)
    # ====== compute classes' average ====== #
    enroll = compute_class_avg(X, y, classes, sorting=True)
    M = X.mean(axis=0).reshape(1, -1)
    self._mean = M
    if self._centering:
      X = X - M
    # ====== WCCN ====== #
    if self._wccn:
      W = compute_wccn(X, y, classes=None, class_avg=enroll) # [feat_dim, feat_dim]
    else:
      W = 1
    self._W = W
    # ====== preprocess ====== #
    # whitening the data
    if self._wccn:
      X = np.dot(X, W)
    # length normalization
    if self._unit_length:
      X = length_norm(X, axis=-1)
    # linear discriminant analysis
    if self._lda is not None:
      self._lda.fit(X, y) # [nb_classes, nb_classes - 1]
    # ====== enroll vecs ====== #
    self._enroll_vecs = self.normalize(enroll, concat=False)
    # ====== max min ====== #
    if self._lda is not None:
      X = self._lda.transform(X)
      X = length_norm(X, axis=-1, ord=2)
    vmin = X.min(0, keepdims=True)
    vmax = X.max(0, keepdims=True)
    self._vmin, self._vmax = vmin, vmax
    return self

  def transform(self, X):
    return self.normalize(X)

class Scorer(BaseEstimator, TransformerMixin, Evaluable):
  """ Scorer

  Parameters
  ----------
  centering : bool (default: True)
    mean normalized the vectors
  wccn : bool (default: True)
    within class covariance normalization
  lda : bool (default: True)
    Linear Discriminant Analysis
  concat : bool (default: False)
    concatenate original vector to the normalized vector
  method : {'cosine', 'svm'}
    method for scoring

  """

  def __init__(self, centering=True, wccn=True, lda=True, concat=False,
               method='cosine', labels=None):
    super(Scorer, self).__init__()
    self._normalizer = VectorNormalizer(
        centering=centering, wccn=wccn, lda=lda, concat=concat)
    self._labels = labels
    method = str(method).lower()
    if method not in ('cosine', 'svm'):
      raise ValueError('`method` must be one of the following: cosine, svm; '
                       'but given: "%s"' % method)
    self._method = method

  # ==================== properties ==================== #
  @property
  def method(self):
    return self._method

  @property
  def feat_dim(self):
    return self._normalizer.feat_dim

  @property
  def labels(self):
    return self._labels

  @property
  def nb_classes(self):
    return len(self._labels)

  @property
  def is_initialized(self):
    return self._normalizer.is_initialized

  @property
  def is_fitted(self):
    return self._normalizer.is_fitted

  @property
  def normalizer(self):
    return self._normalizer

  @property
  def lda(self):
    return self._normalizer.lda

  # ==================== sklearn ==================== #
  def fit(self, X, y):
    # ====== preprocessing ====== #
    if isinstance(X, (tuple, list)):
      X = np.asarray(X)
    if isinstance(y, (tuple, list)):
      y = np.asarray(y)
    # ====== vector normalizer ====== #
    self._normalizer.fit(X, y)
    if self._labels is None:
      if y.ndim >= 2:
        y = np.argmax(y, axis=-1)
      self._labels = np.unique(y)
    # ====== for SVM method ====== #
    if self.method == 'svm':
      X = self._normalizer.transform(X)
      # normalize to [0, 1]
      X = 2 * (X - self._normalizer.vmin) /\
          (self._normalizer.vmax - self._normalizer.vmin) - 1
      self._svm = SVC(C=1, kernel='rbf', gamma='auto', coef0=1,
                      shrinking=True, random_state=0,
                      probability=True, tol=1e-3,
                      cache_size=1e4, class_weight='balanced')
      self._svm.fit(X, y)
      self.predict_proba = self._predict_proba
    return self

  def _predict_proba(self, X):
    if self.method != 'svm':
      raise RuntimeError("`predict_proba` only for 'svm' method")
    return self._svm.predict_proba(self._normalizer.transform(X))

  def predict_log_proba(self, X):
    return self.transform(X)

  def transform(self, X):
    # [nb_samples, nb_classes - 1] (if LDA applied)
    X = self._normalizer.transform(X)
    # ====== cosine scoring ====== #
    if self.method == 'cosine':
      # [nb_classes, nb_classes - 1]
      model_ivectors = self._normalizer.enroll_vecs
      test_ivectors = X
      scores = np.dot(test_ivectors, model_ivectors.T)
    # ====== svm ====== #
    elif self.method == 'svm':
      X = 2 * (X - self._normalizer.vmin) /\
          (self._normalizer.vmax - self._normalizer.vmin) - 1
      scores = self._svm.predict_log_proba(X)
    return scores



from __future__ import absolute_import

import os
import base64
from collections import Mapping
from six import add_metaclass
from abc import ABCMeta, abstractmethod

from odin.fuel import Dataset
from odin.nnet.base import NNOp, VariableDesc
from odin.utils import get_datasetpath, get_file, is_string, as_tuple, is_number

from zipfile import ZipFile, ZIP_DEFLATED


def _validate_shape_dtype(x):
  if not isinstance(x, tuple):
    return False
  if not len(x) == 2:
    return False
  shape, dtype = x
  # check shape
  if not isinstance(shape, tuple) and \
  all(is_number(i) or isinstance(i, type(None)) for i in x):
    return False
  # check dtype
  if not is_string(dtype):
    return False
  return True

@add_metaclass(ABCMeta)
class Model(NNOp):
  """ Model """
  ORIGIN = b'aHR0cHM6Ly9zMy5hbWF6b25hd3MuY29tL2FpLW0wZGVscy8=\n'
  BASE_DIR = get_datasetpath(root='~')

  def __init__(self, **kwargs):
    super(Model, self).__init__(**kwargs)
    input_info = self.get_input_info()
    if not isinstance(input_info, Mapping) or \
    len(input_info) == 0 or \
    not all(is_string(k) and _validate_shape_dtype(v)
            for k, v in input_info.items()):
      raise ValueError("`get_input_info` must return a (length > 0) Mapping "
          "of: 'input-name' -> (shape-tuple, dtype-string), but the "
          "returned value is: %s" % str(input_info))
    # ====== init kwargs_desc ====== #
    for name, (shape, dtype) in input_info.items():
      self._kwargs_desc[name] = VariableDesc(
          shape=shape, name=name, dtype=dtype)

  @abstractmethod
  def get_input_info(self):
    pass

  def get_loaded_param(self, name):
    ds = self.__class__.load_parameters()
    if is_string(name):
      return_1_param = True
    else:
      return_1_param = False
    name = as_tuple(name, t=str)
    if any(n not in ds for n in name):
      raise RuntimeError("Cannot find parameter with name:'%s' from loaded "
          "dataset at path: '%s'" % (name, ds.path))
    params = [ds[n][:] for n in name]
    return params[0] if return_1_param else tuple(params)

  @classmethod
  def load_parameters(clazz):
    # ====== all path ====== #
    name = clazz.__name__ + '.zip'
    path = os.path.join(base64.decodebytes(Model.ORIGIN).decode(), name)
    param_path = get_datasetpath(name=clazz.__name__, override=False)
    zip_path = os.path.join(Model.BASE_DIR, name)
    # ====== get params files ====== #
    if not os.path.exists(param_path) or \
    len(os.listdir(param_path)) == 0:
      get_file(name, origin=path, outdir=Model.BASE_DIR)
      zf = ZipFile(zip_path, mode='r', compression=ZIP_DEFLATED)
      zf.extractall(path=Model.BASE_DIR)
      zf.close()
      # check if proper unzipped
      if not os.path.exists(param_path) or \
      len(os.listdir(param_path)) == 0:
        raise RuntimeError("Zip file at path:%s is not proper unzipped, "
            "cannot find downloaded parameters at path: %s" %
            (zip_path, param_path))
      else:
        os.remove(zip_path)
    # ====== create and return the params dataset ====== #
    ds = Dataset(param_path, read_only=True)
    return ds



from __future__ import print_function, division, absolute_import

from six.moves import builtins

import tensorflow as tf

from odin.autoconfig import randint
from odin.backend.helpers import *

def softplus_inverse(x):
  """Helper which computes the function inverse of `tf.nn.softplus`."""
  return tf.log(tf.expm1(x))

def relu(x, alpha=0., name='ReLu'):
  if alpha == 0.:
    return tf.nn.relu(x, name=name)
  else:
    with tf.variable_scope(name):
      # We can't use 0.5 and 1 for one and half.  as if alpha is a
      # numpy dtype, they will be considered as float64, so would
      # cause upcast to float64.
      alpha = tf.convert_to_tensor(x, name='alpha',
          dtype=x.dtype.base_dtype)
      f1 = 0.5 * (1 + alpha)
      f2 = 0.5 * (1 - alpha)
      return f1 * x + f2 * tf.abs(x)

def elu(x, alpha=1., name="ELu"):
  with tf.variable_scope(name):
    res = tf.nn.elu(x)
    if alpha != 1.:
      alpha = tf.convert_to_tensor(x, name='alpha',
          dtype=x.dtype.base_dtype)
      res = tf.where(x > 0, res, alpha * res)
    return res

def hard_sigmoid(x, name='HardSigmoid'):
  with tf.variable_scope(name):
    slope = tf.constant(0.2, dtype=x.dtype.base_dtype)
    shift = tf.constant(0.5, dtype=x.dtype.base_dtype)
    x = (x * slope) + shift
    x = tf.clip_by_value(x, 0., 1.)
    return x

def antirectify(x, name="AntiRectify"):
  """
  This is the combination of a sample-wise L2 normalization with the
  concatenation of:
      - the positive part of the input
      - the negative part of the input
  The result is a tensor of samples that are twice as large as
  the input samples.
  It can be used in place of a ReLU.
      - Input shape: 2D tensor of shape (samples, n)
      - Output shape: 2D tensor of shape (samples, 2*n)

  Notes
  -----
  When applying ReLU, assuming that the distribution of the previous
  output is approximately centered around 0., you are discarding half of
  your input. This is inefficient.
  Antirectifier allows to return all-positive outputs like ReLU, without
  discarding any data.
  Tests on MNIST show that Antirectifier allows to train networks with
  twice less parameters yet with comparable classification accuracy
  as an equivalent ReLU-based network.

  """
  with tf.variable_scope(name):
    if x.shape.ndims != 2:
      raise Exception('This Ops only support 2D input.')
    x -= tf.reduce_mean(x, axis=1, keepdims=True)
    # l2 normalization
    x /= tf.sqrt(tf.reduce_sum(tf.square(x), axis=1, keepdims=True))
    return tf.concat([relu(x, 0), relu(-x, 0)], axis=1)


def randrectify(x, lower=0.3, upper=0.8, shared_axes='auto', name="RandRectify"):
  """ This function is adpated from Lasagne
  Original work Copyright (c) 2014-2015 lasagne contributors
  All rights reserved.
  LICENSE: https://github.com/Lasagne/Lasagne/blob/master/LICENSE

  Applies a randomized leaky rectify activation to x.

  The randomized leaky rectifier was first proposed and used in the Kaggle
  NDSB Competition, and later evaluated in [1]_. Compared to the standard
  leaky rectifier :func:`leaky_rectify`, it has a randomly sampled slope
  for negative input during training, and a fixed slope during evaluation.

  Equation for the randomized rectifier linear unit during training:
  :math:`\\varphi(x) = \\max((\\sim U(lower, upper)) \\cdot x, x)`

  During evaluation, the factor is fixed to the arithmetic mean of `lower`
  and `upper`.

  Parameters
  ----------
  lower : Theano shared variable, expression, or constant
      The lower bound for the randomly chosen slopes.

  upper : Theano shared variable, expression, or constant
      The upper bound for the randomly chosen slopes.

  shared_axes : 'auto', 'all', int or tuple of int
      The axes along which the random slopes of the rectifier units are
      going to be shared. If ``'auto'`` (the default), share over all axes
      except for the second - this will share the random slope over the
      minibatch dimension for dense layers, and additionally over all
      spatial dimensions for convolutional layers. If ``'all'``, share over
      all axes, thus using a single random slope.

   References
  ----------
  .. [1] Bing Xu, Naiyan Wang et al. (2015):
     Empirical Evaluation of Rectified Activations in Convolutional Network,
     http://arxiv.org/abs/1505.00853
  """
  ndims = x.shape.ndims
  # ====== check lower and upper ====== #
  if is_variable(lower):
    add_roles(lower, ActivationParameter)
  if is_variable(upper):
    add_roles(upper, ActivationParameter)
  if not is_tensor(lower > upper) and lower > upper:
    raise ValueError("Upper bound for Randomized Rectifier needs "
                     "to be higher than lower bound.")
  # ====== check shared_axes ====== #
  if shared_axes == 'auto':
    shared_axes = (0,) + tuple(range(2, ndims))
  elif shared_axes == 'all':
    shared_axes = tuple(range(ndims))
  elif isinstance(shared_axes, int):
    shared_axes = (shared_axes,)
  else:
    shared_axes = shared_axes
  # ====== main logic ====== #
  if not is_training() or upper == lower:
    x = relu(x, alpha=(upper + lower) / 2.0)
  else: # Training mode
    shape = list(input_shape)
    if builtins.any(s is None for s in shape):
      shape = list(x.shape)
    for ax in shared_axes:
      shape[ax] = 1
    rnd = tf.random_uniform(tuple(shape),
               minval=lower,
               maxval=upper,
               dtype=x.dtype.base_dtype,
               seed=randint())
    x = relu(x, rnd)
  return x



"""The NormalGamma distribution class."""
from __future__ import absolute_import, division, print_function

import math
import functools
# Dependency imports
import numpy as np
import tensorflow as tf

from tensorflow_probability.python.distributions import distribution
from tensorflow_probability.python.distributions import seed_stream

from tensorflow_probability.python.internal import distribution_util
from tensorflow_probability.python.internal import dtype_util
from tensorflow_probability.python.internal import reparameterization
from tensorflow.python.framework import tensor_shape
from tensorflow.python.ops import control_flow_ops

__all__ = [
    "NormalGamma",
]

class NormalGamma(distribution.Distribution):
  """Normal-Gamma distribution.
    [normal gamma](https://en.wikipedia.org/wiki/Normal-gamma_distribution)

  The normal-gamma distribution (or Gaussian-gamma distribution)
  is a bivariate four-parameter family of continuous probability
  distributions.

  It is the conjugate prior of a normal distribution with
  unknown mean and precision.

  """

  def __init__(self,
               loc,
               scale,
               concentration,
               rate,
               validate_args=False,
               allow_nan_stats=True,
               name="NormalGamma"):
    """Initializes a batch of Normal-Gamma distributions.



    Args:
      loc: Floating point tensor; the means of the distribution(s).
      scale: Floating point tensor; the stddevs of the distribution(s).
        Must contain only positive values.
        (`scale = sqrt(lambda)` according to the wikipedia article)
      concentration: Floating point tensor, the concentration params of the
        distribution(s). Must contain only positive values.
      rate: Floating point tensor, the inverse scale params of the
        distribution(s). Must contain only positive values.
      validate_args: Python `bool`, default `False`. When `True` distribution
        parameters are checked for validity despite possibly degrading runtime
        performance. When `False` invalid inputs may silently render incorrect
        outputs.
      allow_nan_stats: Python `bool`, default `True`. When `True`, statistics
        (e.g., mean, mode, variance) use the value "`NaN`" to indicate the
        result is undefined. When `False`, an exception is raised if one or more
        of the statistic's batch members are undefined.
      name: Python `str` name prefixed to Ops created by this class.

    Raises:
      TypeError: if `concentration` and `rate` are different dtypes.
    """
    parameters = dict(locals())
    with tf.name_scope(
        name, values=[loc, scale, concentration, rate]):
      dtype = dtype_util.common_dtype(
          [loc, scale, concentration, rate],
          preferred_dtype=tf.float32)
      loc = tf.convert_to_tensor(loc, name="loc", dtype=dtype)
      scale = tf.convert_to_tensor(
          scale, name="scale", dtype=dtype)
      concentration = tf.convert_to_tensor(
          concentration, name="concentration", dtype=dtype)
      rate = tf.convert_to_tensor(rate, name="rate", dtype=dtype)

      with tf.control_dependencies([
          tf.assert_positive(scale),
          tf.assert_positive(concentration),
          tf.assert_positive(rate),
      ] if validate_args else []):
        self._loc = tf.identity(loc)
        self._scale = tf.identity(scale)
        self._concentration = tf.identity(concentration)
        self._rate = tf.identity(rate)
        tf.assert_same_float_dtype(
            [self._loc, self._scale,
             self._concentration, self._rate])
      # the coefficient for the precision
      self._lambda = tf.square(self._scale)
    super(NormalGamma, self).__init__(
        dtype=self._loc.dtype,
        validate_args=validate_args,
        allow_nan_stats=allow_nan_stats,
        reparameterization_type=reparameterization.FULLY_REPARAMETERIZED,
        parameters=parameters,
        graph_parents=[self._loc, self._scale,
                       self._concentration, self._rate],
        name=name)

  @property
  def loc(self):
    """Distribution parameter for the mean."""
    return self._loc

  @property
  def scale(self):
    """Distribution parameter for the standard deviation."""
    return self._scale

  @property
  def concentration(self):
    """Concentration parameter."""
    return self._concentration

  @property
  def rate(self):
    """Rate parameter."""
    return self._rate

  # ******************** Shape and sampling ******************** #

  def _batch_shape_tensor(self):
    tensors = [self.loc, self.scale,
               self.concentration, self.rate]
    return functools.reduce(tf.broadcast_dynamic_shape,
                            [tf.shape(tensor) for tensor in tensors])

  def _batch_shape(self):
    tensors = [self.loc, self.scale,
               self.concentration, self.rate]
    return functools.reduce(tf.broadcast_static_shape,
                            [tensor.shape for tensor in tensors])

  def _event_shape_tensor(self):
    return tf.constant([2], dtype=tf.int32)

  def _event_shape(self):
    return tensor_shape.vector(2)

  def _sample_n(self, n, seed=None):
    seed = seed_stream.SeedStream(seed, "normal_gamma")
    shape = tf.concat([[n], self.batch_shape_tensor()], 0)

    precision = tf.random_gamma(
        shape=shape,
        alpha=self.concentration,
        beta=self.rate,
        dtype=self.dtype,
        seed=seed())

    scale = tf.sqrt(1 / (self._lambda * precision))
    mean = tf.random_normal(
        shape=shape, mean=0., stddev=1., dtype=self.loc.dtype, seed=seed())
    mean = mean * scale + self.loc

    return tf.concat((tf.expand_dims(mean, axis=-1),
                      tf.expand_dims(precision, axis=-1)), axis=-1)

  # ******************** Log probability ******************** #

  def _log_prob(self, x):
    return self._log_unnormalized_prob(x) - self._log_normalization()

  def _log_unnormalized_prob(self, x):
    mean = tf.squeeze(tf.gather(x, [0], axis=-1), axis=-1)
    precision = self._maybe_assert_valid_sample(
        tf.squeeze(tf.gather(x, [1], axis=-1), axis=-1))
    return (tf.math.xlogy(self.concentration - 0.5, precision)
            - self.rate * precision
            - 0.5 * self._lambda * precision * tf.square(mean - self.loc))

  def _log_normalization(self):
    return (tf.lgamma(self.concentration) + 0.5 * math.log(2. * math.pi)
            - self.concentration * tf.log(self.rate)
            - 0.5 * tf.log(self._lambda))

  # ******************** Moments ******************** #

  def _mean(self):
    mean = self.loc * tf.ones_like(self.scale)
    precision = (self.concentration / self.rate) * tf.ones_like(mean)
    return tf.concat((tf.expand_dims(mean, axis=-1),
                      tf.expand_dims(precision, axis=-1)), axis=-1)

  def _maybe_assert_valid_sample(self, x):
    tf.assert_same_float_dtype(tensors=[x], dtype=self.dtype)
    if not self.validate_args:
      return x
    return control_flow_ops.with_dependencies([
        tf.assert_positive(x),
    ], x)



from __future__ import absolute_import, division, print_function

import collections
import math

import numpy as np
import tensorflow as tf

from tensorflow.contrib.rnn import (RNNCell, MultiRNNCell,
    DropoutWrapper, EmbeddingWrapper,
    InputProjectionWrapper, OutputProjectionWrapper,
    PhasedLSTMCell, GridLSTMCell, GLSTMCell,
    IntersectionRNNCell, UGRNNCell, NASCell,
    LayerNormBasicLSTMCell, HighwayWrapper,
    AttentionCellWrapper, BidirectionalGridLSTMCell,
    TimeFreqLSTMCell, CoupledInputForgetGateLSTMCell)
from tensorflow.python.ops.rnn_cell_impl import (GRUCell, LSTMCell, BasicLSTMCell)
from tensorflow.python.util import nest
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn_ops

from odin.backend.helpers import get_all_variables, get_value
from odin.autoconfig import get_session


class BasicRNNCell(RNNCell):
  """The most basic RNN cell.

  Args:
    num_units: int, The number of units in the RNN cell.
    activation: Nonlinearity to use.  Default: `tanh`.
    reuse: (optional) Python boolean describing whether to reuse variables
     in an existing scope.  If not `True`, and the existing scope already has
     the given variables, an error is raised.
  """

  def __init__(self, num_units,
               kernel_initializer=None,
               bias_initializer=tf.constant_initializer(value=0.),
               activation=None,
               reuse=None):
    super(BasicRNNCell, self).__init__(_reuse=reuse)
    self._num_units = num_units
    self._activation = activation or tf.nn.tanh
    self._bias_initializer = bias_initializer
    self._kernel_initializer = kernel_initializer

  @property
  def state_size(self):
    return self._num_units

  @property
  def output_size(self):
    return self._num_units

  def call(self, inputs, state):
    """Most basic RNN: output = new_state = act(W * input + U * state + B)."""
    from tensorflow.python.ops.rnn_cell_impl import _linear
    output = self._activation(
        _linear([inputs, state], output_size=self._num_units,
        bias=False if self._bias_initializer is None else True,
        kernel_initializer=self._kernel_initializer,
        bias_initializer=self._bias_initializer))
    return output, output


class AttentionCell(RNNCell):
  """Basic attention cell wrapper.

  Implementation based on https://arxiv.org/abs/1409.0473.
  """

  def __init__(self, cell, attn_length, attn_size=None, attn_vec_size=None,
               input_size=None, reuse=None):
    """Create a cell with attention.

    Paramters
    ---------
    cell: an RNNCell, an attention is added to it.
    attn_length: integer, the size of an attention window.
    attn_size: integer, the size of an attention vector. Equal to
        cell.output_size by default.
    attn_vec_size: integer, the number of convolutional features calculated
        on attention state and a size of the hidden layer built from
        base cell state. Equal attn_size to by default.
    input_size: integer, the size of a hidden linear layer,
        built from inputs and attention. Derived from the input tensor
        by default.
    state_is_tuple: If True, accepted and returned states are n-tuples, where
      `n = len(cells)`.  By default (False), the states are all
      concatenated along the column axis.
    reuse: (optional) Python boolean describing whether to reuse variables
      in an existing scope.  If not `True`, and the existing scope already has
      the given variables, an error is raised.

    Return
    ------
    output, (new_state, new_attns, new_attn_states)

    Raises
    ------
      TypeError: if cell is not an RNNCell.
      ValueError: if cell returns a state tuple but the flag
          `state_is_tuple` is `False` or if attn_length is zero or less.
    """
    super(AttentionCell, self).__init__(_reuse=reuse)
    if not isinstance(cell, RNNCell):  # pylint: disable=protected-access
      raise TypeError("The parameter cell is not RNNCell.")
    if attn_length <= 0:
      raise ValueError("attn_length should be greater than zero, got %s"
                       % str(attn_length))
    if attn_size is None:
      attn_size = cell.output_size
    if attn_vec_size is None:
      attn_vec_size = attn_size
    self._cell = cell
    self._attn_vec_size = attn_vec_size
    self._input_size = input_size
    self._attn_size = attn_size
    self._attn_length = attn_length
    self._reuse = reuse

  @property
  def state_size(self):
    size = (self._cell.state_size, self._attn_size,
            self._attn_size * self._attn_length)
    return size

  @property
  def output_size(self):
    return self._attn_size

  def call(self, inputs, state):
    """Long short-term memory cell with attention (LSTMA)."""
    state, attns, attn_states = state
    attn_states = array_ops.reshape(attn_states,
                                    [-1, self._attn_length, self._attn_size])
    input_size = self._input_size
    if input_size is None:
      input_size = inputs.shape.as_list()[1]
    inputs = _linear([inputs, attns], input_size, True)
    lstm_output, new_state = self._cell(inputs, state)
    new_state_cat = array_ops.concat(nest.flatten(new_state), 1)
    new_attns, new_attn_states = self._attention(new_state_cat, attn_states)
    with tf.variable_scope("attn_output_projection"):
      output = _linear([lstm_output, new_attns], self._attn_size, True)
    new_attn_states = array_ops.concat(
        [new_attn_states, array_ops.expand_dims(output, 1)], 1)
    new_attn_states = array_ops.reshape(
        new_attn_states, [-1, self._attn_length * self._attn_size])
    new_state = (new_state, new_attns, new_attn_states)
    return output, new_state

  def _attention(self, query, attn_states):
    conv2d = nn_ops.conv2d
    reduce_sum = math_ops.reduce_sum
    softmax = nn_ops.softmax
    tanh = math_ops.tanh

    with tf.variable_scope("attention"):
      k = tf.get_variable(
          "attn_w", [1, 1, self._attn_size, self._attn_vec_size])
      v = tf.get_variable("attn_v", [self._attn_vec_size])
      hidden = array_ops.reshape(attn_states,
                                 [-1, self._attn_length, 1, self._attn_size])
      hidden_features = conv2d(hidden, k, [1, 1, 1, 1], "SAME")
      y = _linear(query, self._attn_vec_size, True)
      y = array_ops.reshape(y, [-1, 1, 1, self._attn_vec_size])
      s = reduce_sum(v * tanh(hidden_features + y), [2, 3])
      a = softmax(s)
      d = reduce_sum(
          array_ops.reshape(a, [-1, self._attn_length, 1, 1]) * hidden, [1, 2])
      new_attns = array_ops.reshape(d, [-1, self._attn_size])
      new_attn_states = array_ops.slice(attn_states, [0, 1, 0], [-1, -1, -1])
      return new_attns, new_attn_states
